#
# AstroImage.py -- Abstraction of an astronomical data image.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import traceback
from collections import OrderedDict

import numpy as np

from ginga.util import wcs, wcsmod
from ginga.BaseImage import BaseImage, ImageError, Header


class AstroHeader(Header):
    pass


class AstroImage(BaseImage):
    """
    Abstraction of an astronomical data (image).

    NOTE: this module is NOT thread-safe!
    """
    # class variables for WCS and IO can be set
    wcsClass = None
    ioClass = None

    @classmethod
    def set_wcsClass(cls, klass):
        cls.wcsClass = klass

    @classmethod
    def set_ioClass(cls, klass):
        cls.ioClass = klass

    def __init__(self, data_np=None, metadata=None, logger=None,
                 name=None, wcsclass=None, ioclass=None, **kwargs):

        BaseImage.__init__(self, data_np=data_np, metadata=metadata,
                           logger=logger, name=name)

        # wcsclass specifies a pluggable WCS module
        if wcsclass is None:
            if self.wcsClass is None:
                wcsclass = wcsmod.WCS
            else:
                wcsclass = self.wcsClass
        self.wcs = wcsclass(self.logger)

        # ioclass specifies a pluggable IO module
        if ioclass is None:
            if self.ioClass is None:
                from ginga.util.io import io_fits
                ioclass = io_fits.fitsLoaderClass
            else:
                ioclass = self.ioClass

        self.io = ioclass(self.logger)

        if metadata is not None:
            header = self.get_header()
            self.wcs.load_header(header)

        # For navigating multidimensional data
        self.naxispath = []
        self.revnaxis = []
        self._md_data = None

    def setup_data(self, data, naxispath=None):
        # initialize data attribute to something reasonable
        if data is None:
            data = np.zeros((0, 0))
        elif 0 in data.shape:
            data = np.zeros((0, 0))

        # this is a handle to the full data array
        self._md_data = data
        self.axisdim = data.shape

        # this will get reset in set_naxispath() if array is
        # multidimensional
        self._data = self._md_data

        if naxispath is None:
            naxispath = []

        # Set naxispath to drill down to first 2D data slice
        if len(naxispath) == 0:
            naxispath = ([0] * (len(data.shape) - 2))

        self.set_naxispath(naxispath)

    def load_hdu(self, hdu, fobj=None, naxispath=None,
                 **kwargs):
        """NOTE: hdu type must be compatible with ioclass set in this
        AstroImage; for example:
        - astropy loaded HDU with AstropyFitsFileHandler, or
        - fitsio loaded HDU with FitsioFileHandler
        """

        if self.io is None:
            raise ImageError("No IO loader defined")

        self.io.load_hdu(hdu, dstobj=self, fobj=fobj, naxispath=naxispath,
                         **kwargs)

    def load_nddata(self, ndd, naxispath=None):
        """Load from an astropy.nddata.NDData object.
        """
        self.clear_metadata()

        # Make a header based on any NDData metadata
        ahdr = self.get_header()
        ahdr.update(ndd.meta)

        self.setup_data(ndd.data, naxispath=naxispath)

        if ndd.wcs is None:
            # no wcs in ndd obj--let's try to make one from the header
            self.wcs = wcsmod.WCS(logger=self.logger)
            self.wcs.load_header(ahdr)
        else:
            # already have a valid wcs in the ndd object
            # we assume it needs an astropy compatible wcs
            wcsinfo = wcsmod.get_wcs_class('astropy')
            self.wcs = wcsinfo.wrapper_class(logger=self.logger)
            self.wcs.load_nddata(ndd)

    def load_file(self, filespec, **kwargs):

        if self.io is None:
            raise ImageError("No IO loader defined")

        self.io.load_file(filespec, dstobj=self, **kwargs)

    def load_data(self, data_np, naxispath=None, metadata=None):

        self.clear_metadata()

        self.setup_data(data_np, naxispath=naxispath)

        if metadata is not None:
            self.update_metadata(metadata)

    def load_buffer(self, buf, dims, dtype, byteswap=False,
                    naxispath=None, metadata=None):
        data = np.frombuffer(buf, dtype=dtype)
        if byteswap:
            data.byteswap(True)
        data = data.reshape(dims)

        self.load_data(data, naxispath=naxispath, metadata=metadata)

    def get_mddata(self):
        return self._md_data

    def set_naxispath(self, naxispath):
        """Choose a slice out of multidimensional data.
        """
        revnaxis = list(naxispath)
        revnaxis.reverse()

        # construct slice view and extract it
        ndim = min(self.ndim, 2)
        view = tuple(revnaxis + [slice(None)] * ndim)
        data = self.get_mddata()[view]

        if len(data.shape) not in (1, 2):
            raise ImageError(
                "naxispath does not lead to a 1D or 2D slice: {}".format(naxispath))

        self.naxispath = naxispath
        self.revnaxis = revnaxis

        self.set_data(data)

    def set_wcs(self, wcs):
        self.wcs = wcs

    def set_io(self, io):
        self.io = io

    def get_data_size(self):
        return self.get_size()

    def get_header(self, create=True, include_primary_header=None):
        # By convention, the fits header is stored in a dictionary
        # under the metadata keyword 'header'
        if 'header' not in self:
            if not create:
                # TODO: change to ValueError("No header found")
                raise KeyError('header')

            hdr = AstroHeader()
            self.set(header=hdr)
        else:
            hdr = self['header']

        if include_primary_header is None:
            include_primary_header = self.get('inherit_primary_header', False)

        # If it was saved during loading, by convention, the primary fits
        # header is stored in a dictionary under the keyword 'primary_header'
        if include_primary_header and self.has_primary_header():
            # Inherit PRIMARY header for display but keep metadata intact
            displayhdr = AstroHeader()
            displayhdr.merge(hdr)
            primary_hdr = self['primary_header']
            displayhdr.merge(primary_hdr, override_keywords=False)
        else:
            # Normal, separate header
            displayhdr = hdr

        return displayhdr

    def get_keyword(self, kwd, *args):
        """Get an item from the fits header, if any."""
        try:
            kwds = self.get_header()
            return kwds[kwd]
        except KeyError:
            # return a default if there is one
            if len(args) > 0:
                return args[0]
            raise KeyError(kwd)

    def get_keywords_list(self, *args):
        return list(map(self.get_keyword, args))

    def set_keyword(self, kwd, value, create=True):
        kwds = self.get_header(create=create)
        kwd = kwd.upper()
        if not create:
            prev = kwds[kwd]  # noqa, this raises KeyError
        kwds[kwd] = value

    def update_keywords(self, key_dict):
        hdr = self.get_header()
        # Upcase all keywords
        for kwd, val in key_dict.items():
            hdr[kwd.upper()] = val

        # Try to make a wcs object on the header
        if hasattr(self, 'wcs') and self.wcs is not None:
            self.wcs.load_header(hdr)

    def set_keywords(self, **kwds):
        """Set an item in the fits header, if any."""
        return self.update_keywords(kwds)

    def update_metadata(self, key_dict):
        self.update(key_dict)

        # refresh the WCS
        if hasattr(self, 'wcs') and self.wcs is not None:
            header = self.get_header()
            self.wcs.load_header(header)

    def has_primary_header(self):
        primary_hdr = self.get('primary_header', None)
        return isinstance(primary_hdr, AstroHeader)

    def clear_all(self):
        # clear metadata and data
        super(AstroImage, self).clear_all()

        # unreference full data array
        self._md_data = self._data

    def transfer(self, other, astype=None):
        data = self._get_data()
        other.set_data(data.copy(), astype=astype)
        other.update_metadata(self.metadata)

    def copy(self, astype=None):
        data = self._get_data()
        other = AstroImage(data, logger=self.logger)
        self.transfer(other, astype=astype)
        return other

    def as_nddata(self, nddata_class=None):
        "Return a version of ourself as an astropy.nddata.NDData object"
        if nddata_class is None:
            from astropy.nddata import NDData
            nddata_class = NDData

        # transfer header, preserving ordering
        ahdr = self.get_header()
        header = OrderedDict(ahdr.items())
        data = self.get_mddata()

        wcs = None
        if hasattr(self, 'wcs') and self.wcs is not None:
            # for now, assume self.wcs wraps an astropy wcs object
            wcs = self.wcs.wcs

        ndd = nddata_class(data, wcs=wcs, meta=header)
        return ndd

    def as_hdu(self):
        "Return a version of ourself as an astropy.io.fits.PrimaryHDU object"
        from astropy.io import fits

        # transfer header, preserving ordering
        ahdr = self.get_header()
        header = fits.Header(ahdr.items())
        data = self.get_mddata()

        hdu = fits.PrimaryHDU(data=data, header=header)
        return hdu

    def astype(self, type_name):
        """Convert AstroImage object to some other kind of object.
        """
        if type_name == 'nddata':
            return self.as_nddata()

        if type_name == 'hdu':
            return self.as_hdu()

        raise ValueError("Unrecognized conversion type '%s'" % (type_name))

    def save_as_file(self, filepath, **kwdargs):
        data = self._get_data()
        header = self.get_header()
        self.io.save_as_file(filepath, data, header, **kwdargs)

    def pixtocoords(self, x, y, system=None, coords='data'):
        args = [x, y] + self.revnaxis
        return self.wcs.pixtocoords(args, system=system, coords=coords)

    def spectral_coord(self, coords='data'):
        args = [0, 0] + self.revnaxis
        return self.wcs.spectral_coord(args, coords=coords)

    def pixtoradec(self, x, y, format='deg', coords='data'):
        args = [x, y] + self.revnaxis
        ra_deg, dec_deg = self.wcs.pixtoradec(args, coords=coords)

        if format == 'deg':
            return ra_deg, dec_deg
        return wcs.deg2fmt(ra_deg, dec_deg, format)

    def radectopix(self, ra_deg, dec_deg, format='deg', coords='data'):
        if format != 'deg':
            # convert coordinates to degrees
            ra_deg = wcs.lon_to_deg(ra_deg)
            dec_deg = wcs.lat_to_deg(dec_deg)
        return self.wcs.radectopix(ra_deg, dec_deg, coords=coords,
                                   naxispath=self.revnaxis)

    def pixtospec(self, x, y):

        # check for cached wavelength array
        waveimg = self.get('wavelen_array', None)
        if waveimg is None:
            # build wavelength array
            # Check that image is 2D
            naxis = self.get_keyword('NAXIS')
            if naxis != 2:
                raise ValueError("Image must be 2-dimensional")

            # Find the spectral axis (using CTYPE)
            spec_axis = None
            for i in range(1, naxis + 1):
                ctype = self.get_keyword(f'CTYPE{i}')
                if 'LAMBDA' in ctype.upper():
                    spec_axis = i
                    break

            if spec_axis is None:
                # Default to axis 1 if not found
                spec_axis = 1

            # Read WCS keywords for spectral axis
            kwds = [f'CRVAL{spec_axis}', f'CDELT{spec_axis}', f'CRPIX{spec_axis}', f'NAXIS{spec_axis}']
            crval, cdelt, crpix, naxis_spec = self.get_keywords_list(*kwds)
            naxis_spat = self.get_keyword(f'NAXIS{3 - spec_axis}')

            # Build wavelength array
            pix_arr = np.arange(naxis_spec) - (crpix - 1)
            wave_array = crval + pix_arr * cdelt

            # Create 2D wavelength image matching data shape
            _ht, _wd = (naxis_spat, naxis_spec) if spec_axis == 1 else (naxis_spec, naxis_spat)  # height, width

            if spec_axis == 1:
                # Spectral axis is along rows
                waveimg = np.tile(wave_array, (_ht, 1))
            elif spec_axis == 2:
                # Spectral axis is along columns
                waveimg = np.tile(wave_array[:, np.newaxis], _wd)
            else:
                raise ValueError(f"Unsupported spectral axis: {spec_axis}")
            self.set(wavelen_array=waveimg)

        _ht, _wd = waveimg.shape

        # We report the value across the pixel, even though the coords
        # change halfway across the pixel
        _x = int(np.floor(x + 0.5))
        _y = int(np.floor(y + 0.5))
        wav = waveimg[_y, _x] if 0 <= _y < _ht and 0 <= _x < _wd else None
        #labels
        wav_lbl, spat_lbl = "\u03bb", ""

        return wav, None, wav_lbl, spat_lbl

    def info_xy(self, data_x, data_y, settings):
        info = super().info_xy(data_x, data_y, settings)

        system = settings.get('wcs_coords', None)
        format = settings.get('wcs_display', 'sexagesimal')
        ra_lbl, dec_lbl = chr(945), chr(948)

        # Calculate WCS coords, if available
        try:
            if self.wcs is None:
                self.logger.debug("No WCS for this image")
                ra_txt = dec_txt = 'NO WCS'

            elif self.wcs.coordsys == 'raw':
                self.logger.debug("No coordinate system determined")
                ra_txt = dec_txt = 'NO WCS'

            elif self.wcs.coordsys == 'pixel':
                args = [data_x, data_y] + self.revnaxis
                x, y = self.wcs.pixtosystem(args, system=system, coords='data')
                ra_txt = "%+.3f" % (x)
                dec_txt = "%+.3f" % (y)
                ra_lbl, dec_lbl = "X", "Y"
            elif self.wcs.coordsys == 'spectral':
                wav, spat, wav_lb, spat_lbl = self.pixtospec(data_x, data_y)
                ra_txt = f"{wav:<14.6g}" if wav is not None else ''
                dec_txt = ""
                ra_lbl, dec_lbl = wav_lb, spat_lbl

            else:
                args = [data_x, data_y] + self.revnaxis

                lon_deg, lat_deg = self.wcs.pixtosystem(
                    args, system=system, coords='data')

                if format == 'sexagesimal':
                    if system in ('galactic', 'ecliptic'):
                        sign, deg, min, sec = wcs.degToDms(lon_deg,
                                                           isLatitude=False)
                        ra_fmt = '+%03d:%02d:%02d.%03d'
                    else:
                        deg, min, sec = wcs.degToHms(lon_deg)
                        ra_fmt = '%02d:%02d:%02d.%03d'

                    frac_sec, sec = np.modf(sec)
                    frac_sec = int(frac_sec * 1000)
                    ra_txt = ra_fmt % (deg, min, sec, frac_sec)

                    sign, deg, min, sec = wcs.degToDms(lat_deg)
                    if sign < 0:
                        sign = '-'
                    else:
                        sign = '+'
                    frac_sec, sec = np.modf(sec)
                    frac_sec = int(frac_sec * 1000)
                    dec_txt = '%s%02d:%02d:%02d.%03d' % (sign, deg, min, sec,
                                                         frac_sec)

                else:
                    ra_txt = '%+10.7f' % (lon_deg)
                    dec_txt = '%+10.7f' % (lat_deg)

                if system == 'galactic':
                    ra_lbl, dec_lbl = "l", "b"
                elif system == 'ecliptic':
                    ra_lbl, dec_lbl = chr(0x03BB), chr(0x03B2)
                elif system == 'helioprojective':
                    ra_txt = "%+5.3f" % (lon_deg * 3600)
                    dec_txt = "%+5.3f" % (lat_deg * 3600)
                    ra_lbl, dec_lbl = "x-Solar", "y-Solar"

        except Exception as e:
            self.logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = dec_txt = 'BAD WCS'
            try:
                # log traceback, if possible
                (type_, value_, tb) = sys.exc_info()
                tb_str = "".join(traceback.format_tb(tb))
                self.logger.error("Traceback:\n%s" % (tb_str))
            except Exception:
                tb_str = "Traceback information unavailable."
                self.logger.error(tb_str)

        info.update(dict(itype='astro', ra_txt=ra_txt, dec_txt=dec_txt,
                         ra_lbl=ra_lbl, dec_lbl=dec_lbl))
        return info



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
Simple data processing pipeline plugin for Ginga.

**Plugin Type: Local**

``Pipeline`` is a local plugin, which means it is associated with a channel.
An instance can be opened for each channel.

**Usage**

"""
import os.path

from ginga import GingaPlugin
from ginga.util import pipeline, paths
from ginga.gw import Widgets

from ginga.util.stages.stage_info import get_stage_catalog

__all__ = ['Pipeline']


class Pipeline(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(Pipeline, self).__init__(fv, fitsimage)

        # Load preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Pipeline')
        self.settings.set_defaults(output_suffix='-pipe')
        self.settings.load(onError='silent')

        self.stage_dict = get_stage_catalog(self.logger)
        self.stage_classes = list(self.stage_dict.values())
        self.stage_names = list(self.stage_dict.keys())
        self.stage_names.sort()

        stages = [self.stage_dict['input'](),
                  self.stage_dict['preview'](),
                  self.stage_dict['output']()]
        self.pipeline = pipeline.Pipeline(self.logger, stages)
        self.pipeline.add_callback('stage-executing', self.stage_status, 'X')
        self.pipeline.add_callback('stage-done', self.stage_status, 'D')
        self.pipeline.add_callback('stage-errored', self.stage_status, 'E')
        self.pipeline.add_callback('pipeline-start', self.clear_status)

        self.pipeline.set(fv=self.fv, viewer=self.fitsimage)
        self.gui_up = False

    def build_gui(self, container):
        top = Widgets.VBox()
        top.set_border_width(4)
        top.set_spacing(2)

        tbar = Widgets.Toolbar(orientation='horizontal')
        menu = tbar.add_menu('Pipe', mtype='menu')
        menu.set_tooltip("Operation on pipeline")
        item = menu.add_name('Load')
        item.set_tooltip("Load a new pipeline")
        item.add_callback('activated', self.load_pipeline_cb)
        item = menu.add_name('Save')
        item.set_tooltip("Save this pipeline")
        item.add_callback('activated', self.save_pipeline_cb)

        menu = tbar.add_menu('Edit', mtype='menu')
        menu.set_tooltip("Edit on pipeline")
        item = menu.add_name('Undo')
        item.set_tooltip("Undo last action")
        item.add_callback('activated', self.undo_pipeline_cb)
        item = menu.add_name('Redo')
        item.set_tooltip("Redo last action")
        item.add_callback('activated', self.redo_pipeline_cb)

        name = Widgets.TextEntry(editable=True)
        name.add_callback('activated', self.set_pipeline_name_cb)
        name.set_text(self.pipeline.name)
        self.w.pipeline_name = name
        tbar.add_widget(name)
        top.add_widget(tbar, stretch=0)

        vbox = Widgets.VBox()
        vbox.set_border_width(2)
        vbox.set_spacing(2)
        self.pipelist = vbox

        for stage in self.pipeline:
            stage_gui = self.make_stage_gui(stage)
            vbox.add_widget(stage_gui, stretch=0)

        # add stretch
        vbox.add_widget(Widgets.Label(''), stretch=1)

        # wrap it in a scrollbox
        scr = Widgets.ScrollArea()
        scr.set_widget(vbox)

        name = self.pipeline.name
        if len(name) > 20:
            name = name[:20] + '...'
        fr = Widgets.Frame("Pipeline: {}".format(name))
        self.w.gui_fr = fr
        fr.set_widget(scr)
        #top.add_widget(scr, stretch=1)
        top.add_widget(fr, stretch=1)

        tbar = Widgets.Toolbar(orientation='horizontal')
        btn = tbar.add_action('Del')
        btn.add_callback('activated', self.delete_stage_cb)
        btn.set_tooltip("Delete selected stages")
        self.w.delete = btn
        btn = tbar.add_action('Ins')
        btn.set_tooltip("Insert above selected stage")
        btn.add_callback('activated', self.insert_stage_cb)
        self.w.insert = btn
        btn = tbar.add_action('Up')
        btn.set_tooltip("Move selected stage up")
        btn.add_callback('activated', self.move_stage_cb, 'up')
        self.w.move_up = btn
        btn = tbar.add_action('Dn')
        btn.set_tooltip("Move selected stage down")
        btn.add_callback('activated', self.move_stage_cb, 'down')
        self.w.move_dn = btn
        btn = tbar.add_action('Clr')
        btn.set_tooltip("Clear selection")
        btn.add_callback('activated', lambda w: self.clear_selected())
        self.w.clear = btn
        self.insert_menu = Widgets.Menu()
        for name in self.stage_names:
            item = self.insert_menu.add_name(name)
            item.add_callback('activated', self._insert_stage_cb, name)
        btn = tbar.add_action('Run')
        btn.set_tooltip("Run entire pipeline")
        btn.add_callback('activated', self.run_pipeline_cb)
        self.w.run = btn
        btn = tbar.add_action('En', toggle=True)
        btn.set_tooltip("Enable pipeline")
        btn.set_state(self.pipeline.enabled)
        btn.add_callback('activated', self.enable_pipeline_cb)
        self.w.enable = btn
        top.add_widget(tbar, stretch=0)

        status = Widgets.Label('')
        self.w.pipestatus = status
        top.add_widget(status, stretch=0)

        btns = Widgets.HBox()
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(btns, stretch=0)

        self._update_toolbar()
        container.add_widget(top, stretch=1)
        self.gui_up = True

    def make_stage_gui(self, stage):
        _vbox = Widgets.VBox()
        hbox = Widgets.HBox()

        xpd = Widgets.Expander(title=str(stage), notoggle=True)
        tbar = Widgets.Toolbar(orientation='horizontal')
        chk = tbar.add_action('B', toggle=True)
        chk.set_state(stage._bypass)
        chk.add_callback('activated', self.bypass_stage_cb, stage)
        chk.set_tooltip("Bypass this stage")
        chk = tbar.add_action('S', toggle=True)
        chk.add_callback('activated', self.select_stage_cb)
        chk.set_tooltip("Select this stage")
        stage.w.select = chk
        chk = tbar.add_action('C', toggle=True)
        chk.add_callback('activated', self.configure_stage_cb, xpd)
        chk.set_tooltip("Configure this stage")
        hbox.add_widget(tbar, stretch=0)
        ent = Widgets.TextEntry(str(stage))
        ent.add_callback('activated', self.rename_stage_cb, stage)
        ent.set_tooltip("Rename this stage")
        hbox.add_widget(ent, stretch=1)
        _vbox.add_widget(hbox, stretch=0)
        stage.build_gui(xpd)
        xpd.add_callback('opened', lambda w: stage.resume())
        xpd.add_callback('closed', lambda w: stage.pause())
        stage.gui_up = True
        _vbox.add_widget(xpd, stretch=0)
        stage.w.gui = _vbox
        return _vbox

    def close(self):
        chname = self.fv.get_channel_name(self.fitsimage)
        self.fv.stop_local_plugin(chname, str(self))
        self.gui_up = False
        return True

    def start(self):
        for stage in self.pipeline:
            stage.start()

        # load any image in the channel into the start of the pipeline
        image = self.fitsimage.get_image()
        if image is not None:
            self.pipeline[0].set_image(image)

    def pause(self):
        pass

    def resume(self):
        pass

    def stop(self):
        self.gui_up = False

        for stage in self.pipeline:
            stage.stop()

    def set_pipeline_name_cb(self, widget):
        name = widget.get_text().strip()
        self.pipeline.name = name
        if len(name) > 20:
            name = name[:20] + '...'
        self.w.gui_fr.set_text("Pipeline: {}".format(name))

    def bypass_stage_cb(self, widget, tf, stage):
        idx = self.pipeline.index(stage)
        stage.bypass(tf)
        self.pipeline.run_from(stage)

    def _update_toolbar(self):
        stages = self.get_selected_stages()
        self.w.insert.set_enabled(len(stages) <= 1)
        self.w.delete.set_enabled(len(stages) >= 1)
        self.w.move_up.set_enabled(len(stages) == 1)
        self.w.move_dn.set_enabled(len(stages) == 1)
        self.w.clear.set_enabled(len(stages) >= 1)
        self.w.run.set_enabled(len(stages) <= 1)
        self.w.enable.set_enabled(len(self.pipeline) > 0)

    def select_stage_cb(self, widget, tf):
        self._update_toolbar()

    def configure_stage_cb(self, widget, tf, xpd):
        xpd.expand(tf)

    def rename_stage_cb(self, widget, stage):
        name = widget.get_text().strip()
        stage.name = name

    def get_selected_stages(self):
        res = [stage for stage in self.pipeline
               if stage.w.select.get_state()]
        return res

    def clear_selected(self):
        for stage in self.pipeline:
            stage.w.select.set_state(False)
        self._update_toolbar()

    def _remove_stage(self, stage, destroy=False):
        self.pipeline.remove(stage)
        self.pipelist.remove(stage.w.gui, delete=destroy)
        if destroy:
            # destroy stage gui
            stage.gui_up = False
            stage.stop()
            stage.w = None

    def delete_stage_cb(self, widget):
        stages = self.get_selected_stages()
        self.clear_selected()
        for stage in stages:
            self._remove_stage(stage, destroy=True)

    def insert_stage_cb(self, widget):
        stages = self.get_selected_stages()
        if len(stages) > 1:
            self.fv.show_error("Please select at most only one stage",
                               raisetab=True)
            return
        self.insert_menu.popup()

    def _insert_stage_cb(self, widget, name):
        stages = self.get_selected_stages()
        if len(stages) == 1:
            stage = stages[0]
            idx = self.pipeline.index(stage)
        else:
            idx = len(stages)
        # realize this stage
        stage = self.stage_dict[name]()
        self.pipeline._init_stage(stage)
        self.pipeline.insert(idx, stage)
        stage_gui = self.make_stage_gui(stage)
        self.pipelist.insert_widget(idx, stage_gui, stretch=0)

    def _relocate_stage(self, idx, stage):
        self._remove_stage(stage, destroy=False)
        self.pipeline.insert(idx, stage)
        self.pipelist.insert_widget(idx, stage.w.gui, stretch=0)

    def move_up(self, stage):
        idx = self.pipeline.index(stage)
        if idx == 0:
            # stage is already at the top
            return
        self._relocate_stage(idx - 1, stage)

    def move_down(self, stage):
        idx = self.pipeline.index(stage)
        if idx == len(self.pipeline) - 1:
            # stage is already at the bottom
            return
        self._relocate_stage(idx + 1, stage)

    def move_stage_cb(self, widget, direction):
        stages = self.get_selected_stages()
        if len(stages) != 1:
            self.fv.show_error("Please select only a single stage",
                               raisetab=True)
        stage = stages[0]
        if direction == 'up':
            self.move_up(stage)
        else:
            self.move_down(stage)

    def run_pipeline_cb(self, widget):
        self.pipeline.run_all()

    def run_pipeline_partial_cb(self, widget):
        stages = self.get_selected_stages()
        if len(stages) == 0:
            self.pipeline.run_all()
            return
        if len(stages) != 1:
            self.fv.show_error("Please select only a single stage",
                               raisetab=True)
            return
        stage = stages[0]
        self.pipeline.run_from(stage)

    def enable_pipeline_cb(self, widget, tf):
        self.pipeline.enable(tf)

    def undo_pipeline_cb(self, widget):
        self.pipeline.undo()

    def redo_pipeline_cb(self, widget):
        self.pipeline.redo()

    def save_pipeline(self, path):
        try:
            d = self.pipeline.save()
        except Exception as e:
            self.logger.error(f"error collecting pipeline configuration: {e}",
                              exc_info=True)
            return
        try:
            import yaml
            with open(path, 'w') as out_f:
                out_f.write(yaml.dump(d))
        except Exception as e:
            self.logger.error(f"error writing pipeline configuration to {path}: {e}",
                              exc_info=True)

    def load_pipeline(self, path):
        self.pipelist.remove_all(delete=True)
        self.pipelist.add_widget(Widgets.Label(''), stretch=1)

        try:
            import yaml
            with open(path, 'r') as in_f:
                s = in_f.read()
            d = yaml.safe_load(s)
            self.pipeline.load(d, self.stage_dict)
        except Exception as e:
            self.logger.error(f"error reading pipeline configuration from {path}: {e}",
                              exc_info=True)
            return

        self.pipeline.set(fv=self.fv, viewer=self.fitsimage)

        for i, stage in enumerate(self.pipeline):
            stage_gui = self.make_stage_gui(stage)
            self.pipelist.insert_widget(i, stage_gui, stretch=0)

        name = self.pipeline.name
        self.w.pipeline_name.set_text(name)
        if len(name) > 20:
            name = name[:20] + '...'
        self.w.gui_fr.set_text(name)

    def save_pipeline_cb(self, widget):
        fname = self.pipeline.name + '.yml'
        save_file = os.path.join(paths.ginga_home, 'pipelines', fname)
        self.save_pipeline(save_file)

    def load_pipeline_cb(self, widget):
        fname = self.w.pipeline_name.get_text() + '.yml'
        save_file = os.path.join(paths.ginga_home, 'pipelines', fname)
        self.load_pipeline(save_file)

    def redo(self):
        image = self.fitsimage.get_image()
        if image is not None:
            stage0 = self.pipeline[0]
            stage0.set_image(image)

    def stage_status(self, pipeline, stage, txt):
        if stage.gui_up:
            self.w.pipestatus.set_text(txt + ': ' + stage.name)
            self.fv.update_pending()

    def clear_status(self, pipeline, stage):
        for stage in pipeline:
            if stage.gui_up:
                self.w.pipestatus.set_text(stage.name)
        self.fv.update_pending()

    def __str__(self):
        return 'pipeline'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Pipeline', package='ginga')

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
# viewer.py -- maintains information about registered viewers for data
#
import uuid
import logging

from ginga.misc import Bunch, Callback, Settings

# this holds registered viewers
viewer_db = {}


class ViewerBase(Callback.Callbacks):
    """An abstract base class for Ginga viewers.
    """

    # subclass should define these two class attributes for viewer name
    # and types that can be viewed
    vname = 'Abstract Ginga Viewer'
    vtypes = []

    @classmethod
    def viewable(cls, dataobj):
        """Test whether `dataobj` is viewable by this viewer.

        Subclass should override this, normally implemented by checking
        whether `dataobj` is an instance of one of the classes listed
        in class variable `vtypes`.
        """
        return False

    def __init__(self, logger=None, settings=None):
        """An abstract base class for Ginga viewers.

        Parameters
        ----------
        logger : :py:class:`~logging.Logger` or `None`
            Logger for tracing and debugging. If not given, one will be created.

        settings : `~ginga.misc.Settings.SettingGroup` or `None`
            Viewer preferences. If not given, one will be created.

        """
        Callback.Callbacks.__init__(self)

        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.Logger('ViewerLogger')

        # Create settings and set defaults
        if settings is None:
            settings = Settings.SettingGroup(logger=self.logger)
        self.settings = settings

        self.allowed_modes = None

        # for debugging
        self.viewer_id = str(uuid.uuid4())
        self.name = self.viewer_id

    def __str__(self):
        return self.name

    def set_name(self, name):
        """Set viewer name."""
        self.name = name

    def get_settings(self):
        """Get the settings used by this instance.

        Returns
        -------
        settings : `~ginga.misc.Settings.SettingGroup`
            Settings.

        """
        return self.settings

    def get_logger(self):
        """Get the logger used by this instance.

        Returns
        -------
        logger : :py:class:`~logging.Logger`
            Logger.

        """
        return self.logger

    def initialize_channel(self, fv, channel):
        """Subclass should override if there is any initialization when
        added to a channel in the reference viewer.
        """
        pass

    def set_allowed_modes(self, mode_list: list or None):
        if mode_list is None:
            mode_list = []
        self.allowed_modes = set(mode_list)

    def is_mode_allowed(self, mode: str):
        return self.allowed_modes is None or mode in self.allowed_modes


def register_viewer(vclass, priority=0):
    """Register a channel viewer.

    Parameters
    ----------
    vclass : :py:class
        The class of the viewer.
    """
    global viewer_db
    viewer_db[vclass.vname] = Bunch.Bunch(name=vclass.vname,
                                          vclass=vclass,
                                          priority=priority)


def get_viewers(dataobj):
    """Returns a list of viewers that are registered that can view `dataobj`,
    sorted by priority.
    """
    res = []
    for vinfo in viewer_db.values():
        vclass = vinfo.vclass
        if not hasattr(vclass, 'viewable'):
            continue
        try:
            tf = vclass.viewable(dataobj)
            if tf:
                res.append(vinfo)
        except Exception as e:
            pass

    res.sort(key=lambda v: v.priority)
    return res


def get_priority_viewers(dataobj):
    """Returns a list of viewers that are registered that can view `dataobj`,
    with all but the best (matching) priority removed.
    """
    res = get_viewers(dataobj)
    if len(res) == 0:
        return res

    # if there is more than one possible viewer, return the list of
    # those that have the best (i.e. lowest) equal priority
    priority = res[0].priority
    return [vinfo for vinfo in res if vinfo.priority <= priority]


def get_viewer_names(dataobj):
    """Returns a list of viewer names that are registered that
    can view `dataobj`.
    """
    return [vinfo.name for vinfo in get_viewers(dataobj)]


def get_vinfo(vname):
    return viewer_db[vname]



try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
except ImportError:
    TESTED_VERSIONS = {}
    PYTEST_HEADER_MODULES = {}

try:
    from ginga import __version__ as version
except ImportError:
    version = 'unknown'

# Uncomment and customize the following lines to add/remove entries
# from the list of packages for which version numbers are displayed
# when running the tests.
PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
PYTEST_HEADER_MODULES['photutils'] = 'photutils'
PYTEST_HEADER_MODULES['Pillow'] = 'PIL'
PYTEST_HEADER_MODULES['QtPy'] = 'qtpy'
PYTEST_HEADER_MODULES.pop('h5py', None)
PYTEST_HEADER_MODULES.pop('Pandas', None)

TESTED_VERSIONS['ginga'] = version



#
# Plot.py -- Plotting widget canvas wrapper.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

# GUI imports
from ginga.qtw import Widgets
from ginga.toolkit import toolkit
# NOTE: imported here so available when importing ginga.gw.Plot
from ginga.mplw.EventMixin import PlotEventMixin  # noqa

if toolkit in ('qt6', 'pyside6'):
    from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
else:
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas


class PlotWidget(Widgets.WidgetBase):

    def __init__(self, plot, width=500, height=500):
        super(PlotWidget, self).__init__()

        self.widget = FigureCanvas(plot.get_figure())

        if plot is not None:
            self.set_plot(plot)

    def set_plot(self, plot):
        self.plot = plot
        self.logger = plot.logger
        self.logger.debug("set_plot called")

        plot.connect_ui(self.widget)



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
``Crosshair`` is a simple plugin to draw crosshairs labeled with the
position of the cross in pixels coordinates, WCS coordinates, or
data value at the cross position.

**Plugin Type: Local**

``Crosshair`` is a local plugin, which means it is associated with a channel.
An instance can be opened for each channel.

**Usage**

Select the appropriate type of output in the "Format" drop-down
box in the UI: "xy" for pixel coordinates, "coords" for the WCS
coordinates, and "value" for the value at the crosshair position.

If "Drag only" is checked, then the crosshair is only updated when the
cursor is clicked or dragged in the window.  If unchecked the crosshair
is positioned by simply moving the cursor around the channel viewer window.

The "Cuts" tab contains a profile plot for the vertical and horizontal
cuts represented by the visible box boundary present when "Quick Cuts"
is checked.  This plot is updated in real time as the crosshair is moved.
When "Quick Cuts" is unchecked, the plot is not updated.

The size of the box is determined by the "radius" parameter.

The "Warn Level" control can be used to set a flux level above which a
warning is indicated in the Cuts plot by a yellow line and the background
turning yellow.  The warning is triggered if any value along either
the X or Y cut exceeds the warn level threshold.

The "Alert Level" control is similar, but represented by a red line and the
background turning pink.  The warning is triggered if any value along either
the X or Y cut exceeds the alert level threshold.  Alerts take precedence
over warnings.

Both the "Warn" and "Alert" features can be turned off by simply setting
a blank value.  They are turned off by default.

The cuts plot is interactive, but it really only makes sense to use that
if "Drag only" is checked.  You can press 'x' or 'y' in the plot window
to toggle on and off the autoaxis scaling feature for either axis, and
scroll in the plot to zoom in the X axis (hold Ctrl down while scrolling
to zoom the Y axis).

Crosshair provides a Pick plugin interaction feature: when the crosshair
is over an object you can press 'r' in the channel viewer window to have
the Pick plugin invoked on that particular location.  If a Pick is not
open already on that channel, it will be opened first.

**User Configuration**

"""
import numpy as np

from ginga import GingaPlugin, BaseImage
from ginga.gw import Widgets, Viewers
from ginga.canvas.types import plots as gplots
from ginga.plot.plotaide import PlotAide

__all__ = ['Crosshair']


class Crosshair(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(Crosshair, self).__init__(fv, fitsimage)

        # get Crosshair preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Crosshair')
        self.settings.add_defaults(color='green', orientation=None,
                                   text_color='skyblue',
                                   box_color='aquamarine',
                                   quick_h_cross_color='#7570b3',
                                   quick_v_cross_color='#1b9e77',
                                   quick_cuts=False,
                                   drag_only=False,
                                   warn_level=None,
                                   alert_level=None,
                                   cuts_radius=15)
        self.settings.load(onError='silent')

        self.layertag = 'crosshair-canvas'
        self.xhtag = None

        self.quick_cuts = self.settings.get('quick_cuts', False)
        self.drag_only = self.settings.get('drag_only', False)

        self.dc = fv.get_draw_classes()
        canvas = self.dc.DrawingCanvas()
        canvas.name = 'crosshair-canvas'
        canvas.enable_draw(True)
        canvas.set_drawtype('squarebox', color='cyan', linestyle='dash')
        canvas.set_callback('draw-event', self.draw_cb)
        canvas.add_draw_mode('move', down=self.cur_down,
                             move=self.cur_drag, up=self.cur_drag,
                             hover=self.cur_hover, key=self.key_down)
        canvas.register_for_cursor_drawing(self.fitsimage)
        canvas.set_draw_mode('move')
        canvas.set_surface(self.fitsimage)
        self.canvas = canvas

        # create crosshair
        self.xh = self.dc.Crosshair(0, 0,
                                    color=self.settings.get('color'),
                                    textcolor=self.settings.get('text_color'),
                                    coord='data', format='xy')
        self.canvas.add(self.xh, redraw=False)

        self.cuts_radius = self.settings.get('cuts_radius', 15)
        alpha = 1.0 if self.quick_cuts else 0.0
        self.cuts_box = self.dc.SquareBox(0, 0, self.cuts_radius,
                                          color=self.settings.get('box_color'),
                                          alpha=alpha,
                                          coord='data')
        self.canvas.add(self.cuts_box, redraw=False)

        self.formats = ('xy', 'value', 'coords')
        self.format = 'xy'
        self._wd, self._ht = 400, 300
        self._split_sizes = [self._ht, self._ht]
        self.gui_up = False

    def handleable(self, dataobj):
        """Test whether `dataobj` can be handled by this plugin."""
        if not isinstance(dataobj, BaseImage.BaseImage):
            return False
        shp = list(dataobj.shape)
        if len(shp) < 2:
            return False
        return True

    def build_gui(self, container):
        top = Widgets.VBox()
        top.set_border_width(4)

        box, sw, orientation = Widgets.get_oriented_box(container,
                                                        orientation=self.settings.get('orientation', None))
        box.set_border_width(4)
        box.set_spacing(2)

        paned = Widgets.Splitter(orientation=orientation)
        self.w.splitter = paned

        nb = Widgets.TabWidget(tabpos='bottom')
        self.w.nb1 = nb
        paned.add_widget(Widgets.hadjust(nb, orientation))

        # Cuts profile plot
        ci = Viewers.CanvasView(logger=self.logger)
        width, height = self._wd, self._ht
        ci.set_desired_size(width, height)
        ci.set_background('white')
        ci.set_foreground('black')
        # for debugging
        ci.set_name('cuts_plot')

        # prepare this viewer as a plot viewer
        self.cuts_view = PlotAide(ci)
        self.cuts_view.setup_standard_frame(title="Cuts",
                                            x_title="Line index",
                                            y_title="Pixel value",
                                            warn_y=None, alert_y=None)
        title = self.cuts_view.get_plot_decor('plot_title')
        title.format_label = self._format_cuts_label

        warn_y = self.settings.get('warn_level', None)
        alert_y = self.settings.get('alert_level', None)
        plot_bg = self.cuts_view.get_plot_decor('plot_bg')
        plot_bg.warn_y = warn_y
        plot_bg.alert_y = alert_y

        # add X and Y data sources. Hereafter, we can just update the data
        # sources and call update_plots() whenever we have new X and Y arms
        cname1 = self.settings.get('quick_h_cross_color', '#7570b3')
        self.cuts_xsrc = gplots.XYPlot(name='X', color=cname1, linewidth=2,
                                       x_acc=np.mean, y_acc=np.nanmax)
        cname2 = self.settings.get('quick_v_cross_color', '#1b9e77')
        self.cuts_ysrc = gplots.XYPlot(name='Y', color=cname2, linewidth=2,
                                       x_acc=np.mean, y_acc=np.nanmax)
        self.cuts_view.add_plot(self.cuts_xsrc)
        self.cuts_view.add_plot(self.cuts_ysrc)

        ciw = Viewers.GingaScrolledViewerWidget(viewer=ci)
        ciw.resize(width, height)
        self.cuts_view.configure_scrollbars(ciw)

        nb.add_widget(ciw, title="Cuts")

        fr = Widgets.Frame("Crosshair")

        captions = (('Format:', 'label', 'format', 'combobox'),
                    ('Drag only', 'checkbutton'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)

        combobox = b.format
        for name in self.formats:
            combobox.append_text(name)
        index = self.formats.index(self.format)
        combobox.set_index(index)
        combobox.add_callback('activated', lambda w, idx: self.set_format())

        b.drag_only.set_tooltip("Must click and drag to update")
        b.drag_only.add_callback('activated', self.drag_only_cb)
        b.drag_only.set_state(self.drag_only)

        fr.set_widget(w)
        box.add_widget(fr, stretch=0)

        fr = Widgets.Frame("Quick Cuts")

        captions = (('Quick Cuts', 'checkbutton'),
                    ('Radius:', 'label', 'radius', 'entryset'),
                    ('Warn Level:', 'label', 'warn_y', 'entryset'),
                    ('Alert Level:', 'label', 'alert_y', 'entryset'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)

        b.quick_cuts.set_tooltip("Do quick cuts plots")
        b.quick_cuts.add_callback('activated', self.quick_cuts_cb)
        b.quick_cuts.set_state(self.quick_cuts)

        b.radius.set_text(str(self.cuts_radius))
        b.radius.set_tooltip("Pixel radius for quick mode cuts")
        b.radius.add_callback('activated', self.set_radius_cb)

        b.warn_y.set_tooltip("Warning level for quick mode cuts")
        b.alert_y.set_tooltip("Alert level for quick mode cuts")
        b.warn_y.set_text('' if warn_y is None else str(warn_y))
        b.alert_y.set_text('' if alert_y is None else str(alert_y))
        b.warn_y.add_callback('activated', self.set_quick_warn_cb)
        b.alert_y.add_callback('activated', self.set_quick_alert_cb)

        fr.set_widget(w)
        box.add_widget(fr, stretch=0)

        spacer = Widgets.Label('')
        box.add_widget(spacer, stretch=1)

        paned.add_widget(sw)
        paned.set_sizes(self._split_sizes)

        top.add_widget(paned, stretch=5)

        btns = Widgets.HBox()
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(btns, stretch=0)

        container.add_widget(top, stretch=1)

        self.gui_up = True

    def set_format(self):
        index = self.w.format.get_index()
        self.format = self.formats[index]
        self.xh.format = self.format

        self.canvas.redraw(whence=3)
        return True

    def close(self):
        self.fv.stop_local_plugin(self.chname, str(self))
        return True

    def start(self):
        # start crosshair operation
        p_canvas = self.fitsimage.get_canvas()
        if not p_canvas.has_object(self.canvas):
            p_canvas.add(self.canvas, tag=self.layertag)

        self.resume()

    def pause(self):
        self.canvas.ui_set_active(False)

    def resume(self):
        self.canvas.ui_set_active(True, viewer=self.fitsimage)
        self.fv.show_status("Click and drag to position crosshair")

    def stop(self):
        self.gui_up = False
        self._split_sizes = self.w.splitter.get_sizes()
        # remove the canvas from the image
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.delete_object_by_tag(self.layertag)
        except Exception:
            pass
        self.canvas.ui_set_active(False)
        self.fv.show_status("")

    def cuts_quick(self, data_x, data_y, radius):
        vip_img = self.fitsimage.get_vip()

        # Get points on the lines
        x0, y0, xarr, yarr = vip_img.cutout_cross(data_x, data_y, radius)

        # plot horizontal cut
        xpts = np.arange(len(xarr))
        points = np.array((xpts, xarr)).T
        self.cuts_xsrc.plot(points)

        # plot vertical cut
        ypts = np.arange(len(yarr))
        points = np.array((ypts, yarr)).T
        self.cuts_ysrc.plot(points)

        self.cuts_view.update_plots()

    def redo(self):
        data_x, data_y = self.xh.x, self.xh.y
        self.cuts_quick(data_x, data_y, self.cuts_radius)

    def move_crosshair(self, viewer, data_x, data_y):
        self.logger.debug("move crosshair data x,y=%f,%f" % (data_x, data_y))
        self.xh.move_to_pt((data_x, data_y))

        if self.quick_cuts:
            self.cuts_box.move_to_pt((data_x, data_y))
            self.cuts_quick(data_x, data_y, self.cuts_radius)

        self.canvas.update_canvas(whence=3)

    def cur_down(self, canvas, event, data_x, data_y, viewer):
        self.move_crosshair(self.fitsimage, data_x, data_y)

    def cur_drag(self, canvas, event, data_x, data_y, viewer):
        self.move_crosshair(self.fitsimage, data_x, data_y)

    def cur_hover(self, canvas, event, data_x, data_y, viewer):
        if not self.drag_only:
            self.move_crosshair(self.fitsimage, data_x, data_y)

    def key_down(self, canvas, event, data_x, data_y, viewer):

        if event.key != 'r':
            return False

        # start Pick plugin on this channel if it hasn't been yet
        opmon = self.channel.opmon
        opname = 'Pick'
        obj = opmon.get_plugin(opname)
        if not opmon.is_active(opname):
            opmon.start_plugin(self.chname, opname)

        # simulate a button down/up in Pick
        self.fv.gui_do(obj.btn_down, canvas, event, data_x, data_y, viewer)
        self.fv.gui_do(obj.btn_up, canvas, event, data_x, data_y, viewer)
        return True

    def draw_cb(self, canvas, tag):
        obj = canvas.get_object_by_tag(tag)
        canvas.delete_object_by_tag(tag)

        self.cuts_radius = int(round(obj.radius))
        self.cuts_box.radius = self.cuts_radius
        # set entry widget to match drawn radius
        self.w.radius.set_text(str(self.cuts_radius))

        self.move_crosshair(self.fitsimage, obj.x, obj.y)

    def drag_only_cb(self, w, tf):
        self.drag_only = tf
        return True

    def quick_cuts_cb(self, w, tf):
        self.quick_cuts = tf
        self.cuts_box.alpha = 1.0 if self.quick_cuts else 0.0
        self.canvas.update_canvas(whence=3)
        return True

    def set_quick_warn_cb(self, w):
        bg = self.cuts_view.get_plot_decor('plot_bg')
        val_s = w.get_text().strip()
        bg.warn_y = None if len(val_s) == 0 else float(val_s)
        # note: necessary to get the warning line updated correctly
        self.cuts_view.update_resize()
        return True

    def set_quick_alert_cb(self, w):
        bg = self.cuts_view.get_plot_decor('plot_bg')
        val_s = w.get_text().strip()
        bg.alert_y = None if len(val_s) == 0 else float(val_s)
        # note: necessary to get the warning line updated correctly
        self.cuts_view.update_resize()
        return True

    def set_radius_cb(self, w):
        radius = float(w.get_text())
        self.cuts_radius = int(round(radius))
        self.cuts_box.radius = self.cuts_radius

        self.move_crosshair(self.fitsimage, self.cuts_box.x, self.cuts_box.y)
        return True

    def _format_cuts_label(self, lbl, plot_src):
        lim = plot_src.get_limits('data')
        y_max = lim[1][1]
        lbl.text = "{0:}: {1: .4g}".format(plot_src.name, y_max)

    def __str__(self):
        return 'crosshair'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Crosshair', package='ginga')



import hashlib
import numpy as np

from ginga import trcalc


class TestTrcalc:

    def _2ddata(self, shape, data_np=None):
        if data_np is None:
            data_np = np.asarray([min(i, j)
                                  for i in range(shape[0])
                                  for j in range(shape[1])])
        data_np = data_np.reshape(shape)
        return data_np

    def _3ddata(self, shape, data_np=None):
        if data_np is None:
            data_np = np.asarray([min(i, j, k)
                                  for i in range(shape[0])
                                  for j in range(shape[1])
                                  for k in range(shape[2])])
        data_np = data_np.reshape(shape)
        return data_np

    def test_get_scaled_cutout_wdht_view(self):

        data = self._2ddata((10, 10))
        p1 = (2, 2)
        p2 = (4, 4)
        nd = (8, 10)

        view, scales = trcalc.get_scaled_cutout_wdht_view(data.shape,
                                                          p1[0], p1[1],
                                                          p2[0], p2[1],
                                                          nd[0], nd[1])
        new_data = trcalc.fancy_index(data, view)
        assert new_data.shape == (10, 8)
        assert isinstance(new_data[0, 0], np.integer)
        res = 'dc025d4e14db5529c581cbe24f0616721bb33f63aabcfcc0d432edf00d8cdc2d'
        m = hashlib.sha256()
        m.update(str(new_data.tolist()).encode())
        assert m.hexdigest() == res

    def test_get_scaled_cutout_wdhtdp_view(self):

        data = self._3ddata((10, 10, 10))
        p1 = (0, 0, 0)
        p2 = (9, 9, 9)
        nd = (4, 4, 4)

        view, scales = trcalc.get_scaled_cutout_wdhtdp_view(data.shape,
                                                            p1, p2, nd)
        new_data = trcalc.fancy_index(data, view)
        assert new_data.shape == (4, 4, 4)
        assert isinstance(new_data[0, 0, 0], np.integer)
        res = 'c01c00af06fb2dc5c8cd6cf96927ba6ddd8d2caba3fc33074c9eaab5cc0ac498'
        m = hashlib.sha256()
        m.update(str(new_data.tolist()).encode())
        assert m.hexdigest() == res



#
# CanvasRenderAgg.py -- for rendering into Ginga widget with aggdraw
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import math
from itertools import chain
import numpy as np

import aggdraw as agg

from . import AggHelp
from ginga.canvas import render
# force registration of all canvas types
import ginga.canvas.types.all  # noqa
from ginga import trcalc


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)

        # TODO: encapsulate this drawable
        self.ctx = AggHelp.AggContext(surface)

        # special scaling for Agg text drawing to normalize it relative
        # to other backends
        self._font_scale_factor = 1.75

    def get_line(self, color, alpha=1.0, linewidth=1, linestyle='solid'):
        line = super().get_line(color, alpha=alpha, linewidth=linewidth,
                                linestyle=linestyle)

        _color = line.get_bpp_color(8)
        line.render.color = _color
        line.render.pen = agg.Pen(_color[:3], width=line.linewidth,
                                  opacity=_color[3])
        return line

    def get_fill(self, color, alpha=1.0):
        fill = super().get_fill(color, alpha=alpha)

        _color = fill.get_bpp_color(8)
        fill.render.color = _color
        fill.render.brush = agg.Brush(_color[:3], opacity=_color[3])
        return fill

    def text_extents(self, text, font=None):
        if font is None:
            font = self.font
        fill = self.get_fill('black')
        _font = self.ctx._get_font(font, fill)
        return self.ctx.text_extents(text, _font)

    def get_affine_transform(self, cx, cy, rot_deg):
        x, y = cx, cy        # old center
        nx, ny = cx, cy      # new center
        sx = sy = 1.0        # new scale
        cosine = math.cos(math.radians(rot_deg))
        sine = math.sin(math.radians(rot_deg))
        a = cosine / sx
        b = sine / sx
        c = x - nx * a - ny * b
        d = -sine / sy
        e = cosine / sy
        f = y - nx * d - ny * e
        return (a, b, c, d, e, f)

    ##### DRAWING OPERATIONS #####

    def draw_image(self, cvs_img, cpoints, rgb_arr, whence, order='RGBA'):
        # no-op for this renderer
        pass

    def draw_text(self, cx, cy, text, rot_deg=0.0, font=None, fill=None,
                  line=None):

        _font = self.ctx._get_font(font, fill)
        wd, ht = self.ctx.text_extents(text, _font)

        affine = self.get_affine_transform(cx, cy, rot_deg)
        self.ctx.canvas.settransform(affine)
        try:
            self.ctx.canvas.text((cx, cy - ht), text, _font)
        finally:
            # reset default transform
            self.ctx.canvas.settransform()

    def draw_polygon(self, cpoints, line=None, fill=None):
        pen = None if line is None else line.render.pen
        brush = None if fill is None else fill.render.brush

        cpoints = trcalc.strip_z(cpoints)
        self.ctx.canvas.polygon(list(chain.from_iterable(cpoints)),
                                pen, brush)

    def draw_circle(self, cx, cy, cradius, line=None, fill=None):
        pen = None if line is None else line.render.pen
        brush = None if fill is None else fill.render.brush

        if line is not None and fill is not None:
            self.ctx.canvas.ellipse(
                (cx - cradius, cy - cradius, cx + cradius, cy + cradius),
                pen, brush)

    # NOTE: recent versions of aggdraw are not rendering Bezier curves
    # correctly--we comment out these temporarily so that we fall back to
    # a version that generates our own Beziers
    #
    # def draw_bezier_curve(self, cp, line=None):
    #     pen = None if line is None else line.render.pen
    #     brush = None if fill is None else fill.render.brush
    #
    #     path = agg.Path()
    #     path.moveto(cp[0][0], cp[0][1])
    #     path.curveto(cp[1][0], cp[1][1], cp[2][0], cp[2][1], cp[3][0], cp[3][1])
    #     self.ctx.canvas.path(path, pen, brush)

    # def draw_ellipse_bezier(self, cp, line=None, fill=None):
    #     # draw 4 bezier curves to make the ellipse because there seems
    #     # to be a bug in aggdraw ellipse drawing function
    #     pen = None if line is None else line.render.pen
    #     brush = None if fill is None else fill.render.brush

    #     path = agg.Path()
    #     path.moveto(cp[0][0], cp[0][1])
    #     path.curveto(cp[1][0], cp[1][1], cp[2][0], cp[2][1], cp[3][0], cp[3][1])
    #     path.curveto(cp[4][0], cp[4][1], cp[5][0], cp[5][1], cp[6][0], cp[6][1])
    #     path.curveto(cp[7][0], cp[7][1], cp[8][0], cp[8][1], cp[9][0], cp[9][1])
    #     path.curveto(cp[10][0], cp[10][1], cp[11][0], cp[11][1], cp[12][0], cp[12][1])
    #     self.ctx.canvas.path(path, pen, brush)

    def draw_line(self, cx1, cy1, cx2, cy2, line=None):
        if line is not None:
            self.ctx.canvas.line((cx1, cy1, cx2, cy2), line.render.pen)

    def draw_path(self, cpoints, line=None):
        if line is not None:
            cp = trcalc.strip_z(cpoints)
            # TODO: is there a more efficient way in aggdraw to do this?
            path = agg.Path()
            path.moveto(cp[0][0], cp[0][1])
            for pt in cp[1:]:
                path.lineto(pt[0], pt[1])
            #brush = agg.Brush(line._color[:3], opacity=line._color[3])
            brush = agg.Brush(line._color[:3], opacity=0)
            self.ctx.canvas.path(path, line.render.pen, brush)


class CanvasRenderer(render.StandardPipelineRenderer):

    def __init__(self, viewer):
        render.StandardPipelineRenderer.__init__(self, viewer)

        self.kind = 'agg'
        self.rgb_order = 'RGBA'
        self.surface = None
        self.dims = ()

    def resize(self, dims):
        """Resize our drawing area to encompass a space defined by the
        given dimensions.
        """
        width, height = dims[:2]
        self.logger.debug("renderer reconfigured to %dx%d" % (
            width, height))
        # create agg surface the size of the window
        self.surface = agg.Draw(self.rgb_order, (width, height), 'black')

        super(CanvasRenderer, self).resize(dims)

    def render_image(self, data, order, win_coord):
        """Render the image represented by (data) at (win_coord)
        in the pixel space.
        *** internal method-- do not use ***
        """
        if self.surface is None:
            return
        self.logger.debug("redraw surface")

        # get window contents as a buffer and load it into the AGG surface
        rgb_buf = data.tobytes(order='C')
        self.surface.frombytes(rgb_buf)

        # for debugging
        # import os.path, tempfile
        # self.save_rgb_image_as_file(os.path.join(tempfile.gettempdir(),
        #                                          'agg_out.png', format='png'))

    def get_surface_as_array(self, order=None):
        if self.surface is None:
            raise render.RenderError("No AGG surface defined")

        # TODO: could these have changed between the time that self.surface
        # was last updated?
        wd, ht = self.dims

        # Get agg surface as a numpy array
        arr8 = np.frombuffer(self.surface.tobytes(), dtype=np.uint8)
        arr8 = arr8.reshape((ht, wd, len(self.rgb_order)))

        # adjust according to viewer's needed order
        return self.reorder(order, arr8)

    def setup_cr(self, shape):
        cr = RenderContext(self, self.viewer, self.surface)
        return cr

    def get_dimensions(self, shape):
        cr = self.setup_cr(shape)
        font = cr.get_font_from_shape(shape)
        return cr.text_extents(shape.text, font=font)

    def text_extents(self, text, font):
        cr = RenderContext(self, self.viewer, self.surface)
        return cr.text_extents(text, font=font)

#END



#
# pan.py -- mode for scaling (zooming) and panning
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import math

from ginga.ImageView import ImageViewBase
from ginga.modes.mode_base import Mode


class PanMode(Mode):
    """
    Pan Mode enables bindings that can set the pan position (the
    position under the center pixel) and zoom level (scale) in a Ginga
    image viewer.

    Enter the mode by
    -----------------
    * Space, then "q"

    Exit the mode by
    ----------------
    * Esc

    Default bindings in mode
    ------------------------
    * plus, equals : zoom in one zoom level
    * minus, underscore : zoom out one zoom level
    * 1-9,0 : zoom to level N (0 is 10)
    * Shift + 1-9,0 : zoom to level -N (0 is -10)
    * backquote : fit image to window size
    * doublequote : toggle auto fit for new images "on" or "off" in this viewer
    * singlequote : set auto fit for new images to "override" in this viewer
    * p : pan to the position under the cursor
    * c : pan to the center of the image
    * z : save zoom level (scale)
    * r : pan to cursor and zoom to saved scale level
    * left/right/up/down arrow : pan left/right/up/down by a small percentage
    * Shift + left/right/up/down arrow : pan left/right/up/down by a
      very small percentage
    * pageup (pagedown) : pan up (down) by a large percentage of the screen
    * home (end) : pan towards the top (bottom) of the image
    * "?" : toggle auto center for new images "on" or "off" in this viewer
    * "/" : set auto center for new images to "override" in this viewer
    * scroll : zoom (scale) the image
    * left drag : pan the view
    * right drag : camera pan the view
    * pan gesture: pan the view porportionally to the gesture
    * pinch gesture: zoom the image porportionally to the gesture

    """

    @classmethod
    def is_compatible_viewer(cls, viewer):
        return isinstance(viewer, ImageViewBase)

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_pan=['__q', None, 'pan'],

            kp_zoom_in=['+', '='],
            kp_zoom_out=['-', '_'],
            kp_zoom=['1', '2', '3', '4', '5', '6', '7', '8', '9', '0'],
            kp_zoom_inv=['!', '@', '#', '$', '%', '^', '&', '*', '(', ')'],
            kp_zoom_fit=['backquote', 'pan+backquote'],
            kp_autozoom_toggle=['doublequote', 'pan+doublequote'],
            kp_autozoom_override=['singlequote', 'pan+singlequote'],
            kp_pan_set=['p', 'pan+p'],
            kp_pan_zoom_set=['pan+r'],
            kp_pan_zoom_save=['pan+z'],
            kp_pan_left=['pan+*+left'],
            kp_pan_right=['pan+*+right'],
            kp_pan_up=['pan+*+up'],
            kp_pan_down=['pan+*+down'],
            kp_pan_home=['pan+*+home'],
            kp_pan_end=['pan+*+end'],
            kp_pan_page_up=['pan+*+page_up'],
            kp_pan_page_down=['pan+*+page_down'],
            kp_pan_px_xminus=['shift+left'],
            kp_pan_px_xplus=['shift+right'],
            kp_pan_px_yminus=['shift+down'],
            kp_pan_px_yplus=['shift+up'],
            kp_pan_px_center=['shift+home'],
            kp_center=['c', 'pan+c'],
            kp_autocenter_toggle=['?', 'pan+?'],
            kp_autocenter_override=['/', 'pan+/'],

            sc_pan=['ctrl+scroll'],
            sc_pan_fine=['pan+shift+scroll'],
            sc_pan_coarse=['pan+ctrl+scroll'],
            sc_zoom=['scroll'],
            sc_zoom_fine=[],
            sc_zoom_coarse=[],
            sc_zoom_origin=['shift+scroll'],

            ms_pan=['pan+left', 'ctrl+left'],
            ms_zoom=['pan+right'],
            ms_panset=['pan+middle', 'shift+left', 'middle'],

            pi_zoom=['pinch'],
            pi_zoom_origin=['shift+pinch'],
            pa_pan=['pan'],

            # can be list of any of: 'zoom', 'rotate'
            pinch_actions=['zoom'],
            scroll_pan_acceleration=1.0,
            scroll_pan_lock_x=False,
            scroll_pan_lock_y=False,
            # 1.0 is appropriate for a mouse, 0.1 for most trackpads
            scroll_zoom_acceleration=1.0,
            #scroll_zoom_acceleration=0.1,
            scroll_zoom_direct_scale=False,

            pan_reverse=False,
            pan_multiplier=1.0,
            #pan_min_scroll_thumb_pct=0.0,
            #pan_max_scroll_thumb_pct=0.9,
            zoom_scroll_reverse=False,

            mouse_zoom_acceleration=1.085,
            # pct of a window of data to move with pan key commands
            key_pan_pct=0.666667,
            # amount to move (in pixels) when using key pan arrow
            key_pan_px_delta=1.0,

            # No messages for following operations:
            msg_panset=False)

        # TODO: define pan cursor

        self._pantype = 1
        self._start_panx = 0
        self._start_pany = 0

        self._start_scale_x = 0
        self._start_scale_y = 0

        self._save = {}

    @property
    def canpan(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('pan')

    @property
    def canzoom(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('zoom')

    @property
    def canrotate(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('rotate')

    def __str__(self):
        return 'pan'

    def start(self):
        self.viewer.switch_cursor('pan')

    def stop(self):
        self.viewer.switch_cursor('pick')
        self.onscreen_message(None)

    #####  Help methods #####
    # Methods used by the callbacks to do actions.

    def get_new_pan(self, viewer, win_x, win_y, ptype=1):

        if ptype == 1:
            # This is a "free pan", similar to dragging the "lens"
            # over the canvas.
            xy_mn, xy_mx = viewer.get_limits()
            win_wd, win_ht = viewer.get_window_size()

            if (win_x >= win_wd):
                win_x = win_wd - 1
            if (win_y >= win_ht):
                win_y = win_ht - 1

            # Figure out data x,y based on percentage of X axis
            # and Y axis
            off_x, off_y = viewer.window_to_offset(win_x, win_y)
            max_x, max_y = viewer.window_to_offset(win_wd, win_ht)
            wd_x = abs(max_x) * 2.0
            ht_y = abs(max_y) * 2.0
            panx = (off_x + abs(max_x)) / float(wd_x)
            pany = (off_y + abs(max_y)) / float(ht_y)

            # Account for user preference
            if self.settings.get('pan_reverse', False):
                panx = 1.0 - panx
                pany = 1.0 - pany

            data_x = (xy_mn[0] + xy_mx[0]) * panx
            data_y = (xy_mn[1] + xy_mx[1]) * pany
            return data_x, data_y

        elif ptype == 2:
            # This is a "drag pan", similar to dragging the canvas
            # under the "lens" or "viewport".
            if self._start_x is None:
                # user has not held the mouse button yet
                # return current pan values
                return (self._start_panx, self._start_pany)

            scale_x, scale_y = viewer.get_scale_xy()
            multiplier = self.settings.get('pan_multiplier', 1.0)
            off_x, off_y = viewer.window_to_offset(win_x, win_y)
            delta_x = float(self._start_x - off_x) / scale_x * multiplier
            delta_y = float(self._start_y - off_y) / scale_y * multiplier

            data_x = self._start_panx + delta_x
            data_y = self._start_pany + delta_y

        return (data_x, data_y)

    def _panset(self, viewer, data_x, data_y, msg=True):
        try:
            msg = self.settings.get('msg_panset', msg)
            if msg:
                self.onscreen_message("Pan position set", delay=0.4)

            res = viewer.panset_xy(data_x, data_y)
            return res

        except Exception as e:
            self.onscreen_message("Pan position set error; see log",
                                  delay=2.0)
            # most likely image does not have a valid wcs
            self.logger.error("Error setting pan position: %s" % (
                str(e)))

    def _get_key_pan_pct(self, event):
        amt = self.settings.get('key_pan_pct', 2 / 3.0)
        if 'ctrl' in event.modifiers:
            amt /= 5.0
        if 'shift' in event.modifiers:
            amt /= 10.0
        return amt

    def _scale_image(self, viewer, direction, factor, msg=True):
        msg = self.settings.get('msg_zoom', msg)
        rev = self.settings.get('zoom_scroll_reverse', False)
        scale_x, scale_y = viewer.get_scale_xy()
        direction = self.get_direction(direction, rev=rev)
        mult = 1.0
        if direction == 'up':
            mult = factor
        elif direction == 'down':
            mult = 1.0 / factor
        scale_x, scale_y = scale_x * mult, scale_y * mult
        scale = max(scale_x, scale_y)
        #viewer.scale_to(scale_x, scale_y)
        viewer.scale_to(scale, scale)
        if msg:
            self.onscreen_message(viewer.get_scale_text(),
                                  delay=0.4)

    def _scale_adjust(self, factor, event_amt, zoom_accel, max_limit=None):
        # adjust scale by factor, amount encoded in event and zoom acceleration value
        amount = factor - ((factor - 1.0) * (1.0 - min(event_amt, 15.0) / 15.0) *
                           zoom_accel)
        amount = max(1.000000001, amount)
        if max_limit is not None:
            amount = min(amount, max_limit)
        return amount

    def _zoom_xy(self, viewer, x, y, msg=True):
        win_wd, win_ht = viewer.get_window_size()
        delta = float(x - self._start_x)
        factor = self.settings.get('mouse_zoom_acceleration', 1.085)
        direction = 0.0
        if delta < 0.0:
            direction = 180.0
        self._start_x = x
        self._scale_image(viewer, direction, factor, msg=msg)

    def _get_pct_xy(self, viewer, x, y):
        win_wd, win_ht = viewer.get_window_size()
        x_pct = float(x - self._start_x) / win_wd
        y_pct = float(y - self._start_y) / win_ht
        return (x_pct, y_pct)

    def to_default_mode(self, viewer):
        self.reset(viewer)

    def pan_start(self, viewer, ptype=1):
        self._pantype = ptype

    def pan_set_origin(self, viewer, win_x, win_y, data_x, data_y):
        self._start_x, self._start_y = viewer.window_to_offset(win_x, win_y)
        self._start_panx, self._start_pany = viewer.get_pan()

    def pan_stop(self, viewer):
        self._start_x = None
        self._pantype = 1

    def zoom_step(self, viewer, event, msg=True, origin=None, adjust=1.5):
        with viewer.suppress_redraw:

            if origin is not None:
                # get cartesian canvas coords of data item under cursor
                data_x, data_y = origin[:2]
                off_x, off_y = viewer.data_to_offset(data_x, data_y)
                # set the pan position to the data item
                viewer.set_pan(data_x, data_y)

            # scale by the desired means
            if self.settings.get('scroll_zoom_direct_scale', True):
                zoom_accel = self.settings.get('scroll_zoom_acceleration', 1.0)
                # change scale by 50%
                amount = self._scale_adjust(adjust, event.amount, zoom_accel,
                                            max_limit=4.0)
                self._scale_image(viewer, event.direction, amount, msg=msg)

            else:
                rev = self.settings.get('zoom_scroll_reverse', False)
                direction = self.get_direction(event.direction, rev=rev)

                if direction == 'up':
                    viewer.zoom_in()

                elif direction == 'down':
                    viewer.zoom_out()

                if msg:
                    self.onscreen_message(viewer.get_scale_text(),
                                          delay=0.4)

            if origin is not None:
                # now adjust the pan position to keep the offset
                data_x2, data_y2 = viewer.offset_to_data(off_x, off_y)
                dx, dy = data_x2 - data_x, data_y2 - data_y
                viewer.panset_xy(data_x - dx, data_y - dy)

    def _sc_zoom(self, viewer, event, msg=True, origin=None):
        if not self.canzoom:
            return

        msg = self.settings.get('msg_zoom', msg)

        self.zoom_step(viewer, event, msg=msg, origin=origin,
                       adjust=1.5)

    def _pinch_zoom_rotate(self, viewer, state, rot_deg, scale, msg=True,
                           origin=None):
        pinch_actions = self.settings.get('pinch_actions', [])

        with viewer.suppress_redraw:

            if state == 'start':
                self._start_scale_x, self._start_scale_y = viewer.get_scale_xy()
                self._start_rot = viewer.get_rotation()
                return

            if origin is not None:
                # get cartesian canvas coords of data item under cursor
                data_x, data_y = origin[:2]
                off_x, off_y = viewer.data_to_offset(data_x, data_y)
                # set the pan position to the data item
                viewer.set_pan(data_x, data_y)

            msg_str = None
            if self.canzoom and 'zoom' in pinch_actions:
                # scale by the desired means
                scale_accel = self.settings.get('pinch_zoom_acceleration', 1.0)
                scale = scale * scale_accel
                # NOTE: scale reported by Qt (the only toolkit that really gives
                # us a pinch gesture right now) the scale reports are iterative
                # so it makes more sense to keep updating the scale than to base
                # the scale on the original scale captured when the gesture starts
                ## scale_x, scale_y = (self._start_scale_x * scale,
                ##                     self._start_scale_y * scale)
                scale_x, scale_y = viewer.get_scale_xy()
                scale_x, scale_y = scale_x * scale, scale_y * scale

                viewer.scale_to(scale_x, scale_y)
                msg_str = viewer.get_scale_text()
                msg = self.settings.get('msg_zoom', True)

            if self.canrotate and 'rotate' in pinch_actions:
                deg = self._start_rot - rot_deg
                rotate_accel = self.settings.get('pinch_rotate_acceleration', 1.0)
                deg = rotate_accel * deg
                viewer.rotate(deg)
                if msg_str is None:
                    msg_str = "Rotate: %.2f" % (deg)
                    msg = self.settings.get('msg_rotate', msg)

            if origin is not None:
                # now adjust the pan position to keep the offset
                data_x2, data_y2 = viewer.offset_to_data(off_x, off_y)
                dx, dy = data_x2 - data_x, data_y2 - data_y
                viewer.panset_xy(data_x - dx, data_y - dy)

            if msg and (msg_str is not None):
                self.onscreen_message(msg_str, delay=0.4)

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_pan_set(self, viewer, event, data_x, data_y, msg=True):
        """Sets the pan position under the cursor."""
        if not self.canpan:
            return False
        event.accept()
        self._panset(viewer, data_x, data_y, msg=msg)

    def kp_pan_zoom_set(self, viewer, event, data_x, data_y, msg=True):
        """Sets the pan position under the cursor."""
        if not self.canpan:
            return False
        event.accept()
        reg = 1
        with viewer.suppress_redraw:
            viewer.panset_xy(data_x, data_y)
            scale_x, scale_y = self._save.get((viewer, 'scale', reg),
                                              (1.0, 1.0))
            viewer.scale_to(scale_x, scale_y)

    def kp_pan_zoom_save(self, viewer, event, data_x, data_y, msg=True):
        """Save the current viewer scale for future use with
        kp_pan_zoom_set()."""
        if not self.canpan:
            return False
        event.accept()
        reg = 1
        scale = viewer.get_scale_xy()
        self._save[(viewer, 'scale', reg)] = scale
        if msg:
            self.onscreen_message("Saved current scale", delay=0.5)

    def kp_pan_left(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        amt = self._get_key_pan_pct(event)
        viewer.pan_lr(amt, -0.1)

    def kp_pan_right(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        amt = self._get_key_pan_pct(event)
        viewer.pan_lr(amt, 0.1)

    def kp_pan_up(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        amt = self._get_key_pan_pct(event)
        viewer.pan_ud(amt, 0.1)

    def kp_pan_down(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        amt = self._get_key_pan_pct(event)
        viewer.pan_ud(amt, -0.1)

    def kp_pan_page_up(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        amt = self._get_key_pan_pct(event)
        pct = 1.0
        if 'ctrl' in event.modifiers:
            # adjust X axis
            viewer.pan_lr(amt, pct)
        else:
            # adjust Y axis
            viewer.pan_ud(amt, pct)

    def kp_pan_page_down(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        amt = self._get_key_pan_pct(event)
        pct = -1.0
        if 'ctrl' in event.modifiers:
            # adjust X axis
            viewer.pan_lr(amt, pct)
        else:
            # adjust Y axis
            viewer.pan_ud(amt, pct)

    def kp_pan_home(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        res = viewer.calc_pan_pct(pad=0)
        # 1.0 == max
        if 'ctrl' in event.modifiers:
            # adjust X axis
            viewer.pan_by_pct(1.0, res.pan_pct_y)
        else:
            # adjust Y axis
            viewer.pan_by_pct(res.pan_pct_x, 1.0)

    def kp_pan_end(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        res = viewer.calc_pan_pct(pad=0)
        # 0.0 == min
        if 'ctrl' in event.modifiers:
            # adjust X axis
            viewer.pan_by_pct(0.0, res.pan_pct_y)
        else:
            # adjust Y axis
            viewer.pan_by_pct(res.pan_pct_x, 0.0)

    def kp_pan_px_xminus(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        px_amt = self.settings.get('key_pan_px_delta', 1.0)
        viewer.pan_delta_px(-px_amt, 0.0)

    def kp_pan_px_xplus(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        px_amt = self.settings.get('key_pan_px_delta', 1.0)
        viewer.pan_delta_px(px_amt, 0.0)

    def kp_pan_px_yminus(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        px_amt = self.settings.get('key_pan_px_delta', 1.0)
        viewer.pan_delta_px(0.0, -px_amt)

    def kp_pan_px_yplus(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        px_amt = self.settings.get('key_pan_px_delta', 1.0)
        viewer.pan_delta_px(0.0, px_amt)

    def kp_pan_px_center(self, viewer, event, data_x, data_y, msg=True):
        """This pans so that the cursor is over the center of the
        current pixel."""
        if not self.canpan:
            return False
        event.accept()
        viewer.pan_center_px()

    def kp_center(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        viewer.center_image()

    def kp_zoom_out(self, viewer, event, data_x, data_y, msg=True):
        if not self.canzoom:
            return False
        event.accept()
        msg = self.settings.get('msg_zoom', msg)
        viewer.zoom_out()
        if msg:
            self.onscreen_message(viewer.get_scale_text(), delay=1.0)

    def kp_zoom_in(self, viewer, event, data_x, data_y, msg=True):
        if not self.canzoom:
            return False
        event.accept()
        msg = self.settings.get('msg_zoom', msg)
        viewer.zoom_in()
        if msg:
            self.onscreen_message(viewer.get_scale_text(), delay=1.0)

    def kp_zoom(self, viewer, event, data_x, data_y, msg=True):
        if not self.canzoom:
            return False
        event.accept()
        msg = self.settings.get('msg_zoom', msg)
        keylist = self.settings.get('kp_zoom')
        try:
            zoomval = (keylist.index(event.key))
        except IndexError:
            return
        viewer.zoom_to(zoomval)
        if msg:
            self.onscreen_message(viewer.get_scale_text(), delay=1.0)

    def kp_zoom_inv(self, viewer, event, data_x, data_y, msg=True):
        if not self.canzoom:
            return False
        event.accept()
        msg = self.settings.get('msg_zoom', msg)
        keylist = self.settings.get('kp_zoom_inv')
        try:
            zoomval = - (keylist.index(event.key))
        except IndexError:
            return
        viewer.zoom_to(zoomval)
        if msg:
            self.onscreen_message(viewer.get_scale_text(), delay=1.0)

    def kp_zoom_fit(self, viewer, event, data_x, data_y, msg=True):
        if not self.canzoom:
            return False
        event.accept()
        msg = self.settings.get('msg_zoom', msg)
        viewer.zoom_fit()
        if msg:
            self.onscreen_message(viewer.get_scale_text(), delay=1.0)

    def kp_autozoom_toggle(self, viewer, event, data_x, data_y, msg=True):
        if not self.canzoom:
            return False
        event.accept()
        msg = self.settings.get('msg_zoom', msg)
        val = viewer.get_settings().get('autozoom')
        if val == 'off':
            val = 'on'
        else:
            val = 'off'
        viewer.enable_autozoom(val)
        if msg:
            self.onscreen_message('Autozoom %s' % val, delay=1.0)

    def kp_autozoom_override(self, viewer, event, data_x, data_y, msg=True):
        if not self.canzoom:
            return False
        event.accept()
        msg = self.settings.get('msg_zoom', msg)
        viewer.enable_autozoom('override')
        if msg:
            self.onscreen_message('Autozoom Override', delay=1.0)

    def kp_autocenter_toggle(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        msg = self.settings.get('msg_pan', msg)
        val = viewer.get_settings().get('autocenter')
        if val == 'off':
            val = 'on'
        else:
            val = 'off'
        viewer.set_autocenter(val)
        if msg:
            self.onscreen_message('Autocenter %s' % val, delay=1.0)

    def kp_autocenter_override(self, viewer, event, data_x, data_y, msg=True):
        if not self.canpan:
            return False
        event.accept()
        msg = self.settings.get('msg_pan', msg)
        viewer.set_autocenter('override')
        if msg:
            self.onscreen_message('Autocenter Override', delay=1.0)

    #####  SCROLL ACTION CALLBACKS #####

    def sc_zoom(self, viewer, event, msg=True):
        """Interactively zoom the image by scrolling motion.
        This zooms by the zoom steps configured under Preferences.
        """
        event.accept()
        self._sc_zoom(viewer, event, msg=msg, origin=None)

    def sc_zoom_origin(self, viewer, event, msg=True):
        """Like sc_zoom(), but pans the image as well to keep the
        coordinate under the cursor in that same position relative
        to the window.
        """
        event.accept()
        origin = (event.data_x, event.data_y)
        self._sc_zoom(viewer, event, msg=msg, origin=origin)

    def sc_zoom_coarse(self, viewer, event, msg=True):
        """Interactively zoom the image by scrolling motion.
        This zooms by adjusting the scale in x and y coarsely.
        """
        if not self.canzoom:
            return False
        event.accept()

        zoom_accel = self.settings.get('scroll_zoom_acceleration', 1.0)
        # change scale by 20%
        amount = self._scale_adjust(1.2, event.amount, zoom_accel, max_limit=4.0)
        self._scale_image(viewer, event.direction, amount, msg=msg)

    def sc_zoom_fine(self, viewer, event, msg=True):
        """Interactively zoom the image by scrolling motion.
        This zooms by adjusting the scale in x and y coarsely.
        """
        if not self.canzoom:
            return False
        event.accept()

        zoom_accel = self.settings.get('scroll_zoom_acceleration', 1.0)
        # change scale by 5%
        amount = self._scale_adjust(1.05, event.amount, zoom_accel, max_limit=4.0)
        self._scale_image(viewer, event.direction, amount, msg=msg)

    def sc_pan(self, viewer, event, msg=True):
        """Interactively pan the image by scrolling motion.
        """
        if not self.canpan:
            return False
        event.accept()

        # User has "Pan Reverse" preference set?
        rev = self.settings.get('pan_reverse', False)
        direction = event.direction

        if rev:
            direction = math.fmod(direction + 180.0, 360.0)

        pan_accel = self.settings.get('scroll_pan_acceleration', 1.0)
        # Internal factor to adjust the panning speed so that user-adjustable
        # scroll_pan_acceleration is normalized to 1.0 for "normal" speed
        scr_pan_adj_factor = 1.4142135623730951
        amount = (event.amount * scr_pan_adj_factor * pan_accel) / 360.0

        lock_x = self.settings.get('scroll_pan_lock_x', False)
        lock_y = self.settings.get('scroll_pan_lock_y', False)

        viewer.pan_omni(direction, amount, lock_x=lock_x, lock_y=lock_y)

    def sc_pan_coarse(self, viewer, event, msg=True):
        if not self.canpan:
            return False
        event.accept()
        event.amount = event.amount / 2.0
        self.sc_pan(viewer, event, msg=msg)

    def sc_pan_fine(self, viewer, event, msg=True):
        if not self.canpan:
            return False
        event.accept()
        event.amount = event.amount / 5.0
        self.sc_pan(viewer, event, msg=msg)

    #####  MOUSE ACTION CALLBACKS #####

    def ms_zoom(self, viewer, event, data_x, data_y, msg=True):
        """Zoom the image by dragging the cursor left or right.
        """
        if not self.canzoom:
            return False
        event.accept()
        msg = self.settings.get('msg_zoom', msg)

        x, y = self.get_win_xy(viewer)

        if event.state == 'move':
            self._zoom_xy(viewer, x, y)

        elif event.state == 'down':
            if msg:
                self.onscreen_message("Zoom (drag mouse L-R)",
                                      delay=1.0)
            self._start_x, self._start_y = x, y

        else:
            self.onscreen_message(None)

    def ms_pan(self, viewer, event, data_x, data_y):
        """A 'drag' or proportional pan, where the image is panned by
        'dragging the canvas' up or down.  The amount of the pan is
        proportionate to the length of the drag.
        """
        if not self.canpan:
            return False
        event.accept()

        x, y = viewer.get_last_win_xy()
        if event.state == 'move':
            data_x, data_y = self.get_new_pan(viewer, x, y,
                                              ptype=self._pantype)
            viewer.panset_xy(data_x, data_y)

        elif event.state == 'down':
            self.pan_set_origin(viewer, x, y, data_x, data_y)
            self.pan_start(viewer, ptype=2)

        else:
            self.pan_stop(viewer)

    def ms_panset(self, viewer, event, data_x, data_y,
                  msg=True):
        """An interactive way to set the pan position.  The location
        (data_x, data_y) will be centered in the window.
        """
        if not self.canpan:
            return False
        event.accept()
        if event.state == 'down':
            self._panset(viewer, data_x, data_y, msg=msg)

    ##### GESTURE ACTION CALLBACKS #####

    def pa_pan(self, viewer, event, msg=True):
        """Interactively pan the image by a pan gesture.
        (the back end must support gestures)
        """
        if not self.canpan:
            return False
        event.accept()

        method = 1
        x, y = viewer.get_last_win_xy()
        dx, dy = event.delta_x, event.delta_y

        # User has "Pan Reverse" preference set?
        rev = self.settings.get('pan_reverse', False)
        if rev:
            dx, dy = -dx, -dy

        if event.state == 'move':
            # Internal factor to adjust the panning speed so that user-adjustable
            # pan_pan_acceleration is normalized to 1.0 for "normal" speed
            pan_pan_adj_factor = 1.0
            pan_accel = (self.settings.get('pan_pan_acceleration', 1.0) *
                         pan_pan_adj_factor)

            if method == 1:
                # METHOD 1
                # similar to moving by the scroll bars
                # calculate current pan pct
                res = viewer.calc_pan_pct(pad=0)

                # modify the pct, as relative to the offsets
                amt_x = float(dx) / res.rng_x * pan_accel
                amt_y = float(dy) / res.rng_y * pan_accel

                pct_x = res.pan_pct_x - amt_x
                pct_y = res.pan_pct_y + amt_y

                # update the pan position by pct
                viewer.pan_by_pct(pct_x, pct_y)

            elif method == 2:
                # METHOD 2
                # similar to using a drag pan
                x, y = x + dx * pan_accel, y + dy * pan_accel

                data_x, data_y = self.get_new_pan(viewer, x, y,
                                                  ptype=self._pantype)
                viewer.panset_xy(data_x, data_y)

            elif method == 3:
                # METHOD 3
                # calculate new position from pan gesture offsets
                data_x, data_y = viewer.get_pan(coord='data')
                x, y = viewer.get_canvas_xy(data_x, data_y)

                x, y = x - dx * pan_accel, y - dy * pan_accel

                data_x, data_y = viewer.get_data_xy(x, y)
                viewer.panset_xy(data_x, data_y)

        elif event.state == 'start':
            #self._start_panx, self._start_pany = viewer.get_pan()
            data_x, data_y = viewer.get_last_data_xy()
            self.pan_set_origin(viewer, x, y, data_x, data_y)
            self.pan_start(viewer, ptype=2)

        else:
            self.pan_stop(viewer)

    def pi_zoom(self, viewer, event, msg=True):
        """Zoom and/or rotate the viewer by a pinch gesture.
        (the back end must support gestures)
        """
        if not self.canzoom:
            return False
        event.accept()
        self._pinch_zoom_rotate(viewer, event.state, event.rot_deg,
                                event.scale, msg=msg)

    def pi_zoom_origin(self, viewer, event, msg=True):
        """Like pi_zoom(), but pans the image as well to keep the
        coordinate under the cursor in that same position relative
        to the window.
        """
        if not self.canzoom:
            return False
        event.accept()
        origin = (event.data_x, event.data_y)
        self._pinch_zoom_rotate(viewer, event.state, event.rot_deg,
                                event.scale, msg=msg, origin=origin)



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga import trcalc
from ginga.gw import Widgets

from .base import Stage, StageAction


class Scale(Stage):
    """
    The Scale stage will scale a 2D image.

    If 'Long Side' (`longside`) is set to some pixel size, then the image
    is scaled so that the long side will be that many pixels; the other
    side will be scaled to keep the aspect ratio.

    If 'Long Side' is not set, or 'Scale' (`scale`) has been set to some
    float value, then the image is scaled according to that scale.

    'Interpolation' (`interp`) informs what kind of interpolation will be
    done to the image during the scaling.
    """

    _stagename = 'scale'

    def __init__(self):
        super().__init__()

        self._longside = None
        self._scale = 1.0
        self._interp = 'lanczos'
        self.viewer = None

    def build_gui(self, container):
        self.viewer = self.pipeline.get('viewer')

        fr = Widgets.Frame("Scale")

        captions = ((),
                    ('Long side:', 'label', 'longside', 'entryset'),
                    ('Interpolation:', 'label', 'Interpolation', 'combobox'),
                    ('Scale:', 'label', 'scale', 'entryset'),
                    ('_sp1', 'spacer', 'hbox1', 'hbox'),
                    ('Output Size:', 'label', 'size', 'llabel'),
                    )
        w, b = Widgets.build_info(captions, orientation='vertical')

        b.longside.set_tooltip("Set the length of the long side in pixels")
        b.longside.set_text("" if self._longside is None
                            else str(self._longside))
        b.longside.add_callback('activated', self.set_longside_cb)

        b.scale.set_tooltip("Set the scale")
        b.scale.set_text(str(self._scale))
        b.scale.add_callback('activated', self.set_scale_cb)

        b.copy_from_viewer = Widgets.Button("Copy from viewer")
        b.copy_from_viewer.set_tooltip("Copy scale setting from viewer")
        b.copy_from_viewer.add_callback('activated', self.copy_from_viewer_cb)
        b.hbox1.add_widget(b.copy_from_viewer, stretch=0)
        b.hbox1.add_widget(Widgets.Label(''), stretch=1)

        index = 0
        for name in trcalc.interpolation_methods:
            b.interpolation.append_text(name)
            index += 1
        try:
            index = trcalc.interpolation_methods.index(self.interp)
        except ValueError:
            # previous choice might not be available if preferences
            # were saved when opencv was being used--if so, default
            # to "basic"
            index = trcalc.interpolation_methods.index('basic')
        b.interpolation.set_index(index)
        b.interpolation.set_tooltip("Choose interpolation method")
        b.interpolation.add_callback('activated', self.set_interp_cb)

        b.size.set_text('unknown')

        self.w.update(b)
        fr.set_widget(w)

        container.set_widget(fr)

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, val):
        self._scale = val
        if self.gui_up:
            self.w.scale.set_text(str(self._scale))

    @property
    def longside(self):
        return self._longside

    @longside.setter
    def longside(self, val):
        if val is not None:
            val = int(val)
        self._longside = val
        if self.gui_up:
            self.w.longside.set_text("" if self._longside is None
                                     else str(self._longside))

    @property
    def interp(self):
        return self._interp

    @interp.setter
    def interp(self, val):
        self._interp = val
        if self.gui_up:
            index = trcalc.interpolation_methods.index(self._interp)
            self.w.interpolation.set_index(index)

    def _get_state(self):
        return dict(scale=self._scale, longside=self._longside,
                    interp=self._interp)

    def _set_scale(self, scale):
        old = self._get_state()
        self.scale = scale
        self.longside = None
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="set scale factor"))
        self.pipeline.run_from(self)

    def set_scale_cb(self, widget):
        scale = float(widget.get_text())
        self._set_scale(scale)

    def _set_longside(self, longside):
        old = self._get_state()
        self.longside = longside
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="set scale longside"))
        self.pipeline.run_from(self)

    def set_longside_cb(self, widget):
        length = widget.get_text().strip()
        if len(length) == 0:
            longside = None
        else:
            longside = int(length)
        self._set_longside(longside)

    def set_interp_cb(self, w, idx):
        old = self._get_state()
        self.interp = trcalc.interpolation_methods[idx]
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="set scale interpolation"))
        self.pipeline.run_from(self)

    def copy_from_viewer_cb(self, widget):
        scale = self.viewer.get_scale()
        self._set_scale(scale)

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        if self._bypass or data is None or (self.longside is None and
                                            np.isclose(self.scale, 1.0)):
            if self.gui_up:
                self.w.scale.set_text(str(1.0))
                _ht, _wd = data.shape[:2]
                self.w.size.set_text("{}x{}".format(_wd, _ht))

            self.pipeline.send(res_np=data)
            return

        ht, wd = data.shape[:2]

        if self.longside is not None:
            aspect = trcalc.get_aspect(data.shape)
            if aspect > 0:
                new_wd, new_ht = (self.longside,
                                  int(np.floor(self.longside / aspect)))
            else:
                new_wd, new_ht = (int(np.floor(self.longside * aspect)),
                                  self.longside)

            res_np, scales = trcalc.get_scaled_cutout_wdht(data, 0, 0,
                                                           wd - 1, ht - 1,
                                                           new_wd, new_ht,
                                                           interpolation=self.interp,
                                                           logger=self.logger)
            if self.gui_up:
                self.w.scale.set_text("%10.4f" % (scales[0]))

        else:
            res_np, scales = trcalc.get_scaled_cutout_basic(data, 0, 0,
                                                            wd - 1, ht - 1,
                                                            self.scale, self.scale,
                                                            interpolation=self.interp,
                                                            logger=self.logger)

        if self.gui_up:
            _ht, _wd = res_np.shape[:2]
            self.w.size.set_text("{}x{}".format(_wd, _ht))

        self.pipeline.send(res_np=res_np)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.scale = d['scale']
        self.longside = d['longside']
        self.interp = d['interp']



#! /usr/bin/env python
#
# example2_qt.py -- Simple, configurable FITS viewer.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

import sys

from ginga.qtw.QtHelp import QtGui, QtCore, set_default_opengl_context
from ginga import colors
from ginga.qtw.ImageViewQt import CanvasView
from ginga.canvas.CanvasObject import get_canvas_types
from ginga.misc import log
from ginga.util.loader import load_data
from ginga.locale.localize import _

STD_FORMAT = '%(asctime)s | %(levelname)1.1s | %(filename)s:%(lineno)d (%(funcName)s) | %(message)s'


class FitsViewer(QtGui.QMainWindow):

    def __init__(self, logger, render='widget'):
        super(FitsViewer, self).__init__()
        self.logger = logger
        self.drawcolors = colors.get_colors()
        self.dc = get_canvas_types()

        fi = CanvasView(logger, render=render)
        fi.enable_autocuts('on')
        fi.set_autocut_params('zscale')
        fi.enable_autozoom('on')
        fi.set_zoom_algorithm('rate')
        fi.set_zoomrate(1.4)
        fi.show_pan_mark(True)
        fi.add_callback('drag-drop', self.drop_file_cb)
        fi.add_callback('cursor-changed', self.cursor_cb)
        fi.set_bg(0.2, 0.2, 0.2)
        fi.ui_set_active(True)
        self.fitsimage = fi

        bd = fi.get_bindings()
        bd.enable_all(True)

        # canvas that we will draw on
        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.set_drawtype('rectangle', color='lightblue')
        canvas.register_for_cursor_drawing(fi)
        canvas.add_callback('draw-event', self.draw_cb)
        canvas.set_draw_mode('draw')
        canvas.set_surface(fi)
        canvas.ui_set_active(True)
        self.canvas = canvas

        # add our new canvas to viewers default canvas
        fi.get_canvas().add(canvas)

        self.drawtypes = canvas.get_drawtypes()
        self.drawtypes.sort()

        # add a color bar
        #fi.show_color_bar(True)
        #fi.show_focus_indicator(True)

        # add little mode indicator that shows keyboard modal states
        fi.show_mode_indicator(True, corner='ur')

        w = fi.get_widget()
        w.resize(512, 512)

        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(QtCore.QMargins(2, 2, 2, 2))
        vbox.setSpacing(1)
        vbox.addWidget(w, stretch=1)

        self.readout = QtGui.QLabel("")
        vbox.addWidget(self.readout, stretch=0,
                       alignment=QtCore.Qt.AlignCenter)

        hbox = QtGui.QHBoxLayout()
        hbox.setContentsMargins(QtCore.QMargins(4, 2, 4, 2))

        wdrawtype = QtGui.QComboBox()
        for name in self.drawtypes:
            wdrawtype.addItem(name)
        index = self.drawtypes.index('rectangle')
        wdrawtype.setCurrentIndex(index)
        wdrawtype.activated.connect(self.set_drawparams)
        self.wdrawtype = wdrawtype

        wdrawcolor = QtGui.QComboBox()
        for name in self.drawcolors:
            wdrawcolor.addItem(name)
        index = self.drawcolors.index('lightblue')
        wdrawcolor.setCurrentIndex(index)
        wdrawcolor.activated.connect(self.set_drawparams)
        self.wdrawcolor = wdrawcolor

        wfill = QtGui.QCheckBox(_("Fill"))
        wfill.stateChanged.connect(self.set_drawparams)
        self.wfill = wfill

        walpha = QtGui.QDoubleSpinBox()
        walpha.setRange(0.0, 1.0)
        walpha.setSingleStep(0.1)
        walpha.setValue(1.0)
        walpha.valueChanged.connect(self.set_drawparams)
        self.walpha = walpha

        wclear = QtGui.QPushButton(_("Clear Canvas"))
        wclear.clicked.connect(self.clear_canvas)
        wopen = QtGui.QPushButton(_("Open File"))
        wopen.clicked.connect(self.open_file)
        wquit = QtGui.QPushButton(_("Quit"))
        wquit.clicked.connect(self.quit)

        hbox.addStretch(1)
        for w in (wopen, wdrawtype, wdrawcolor, wfill,
                  QtGui.QLabel(_('Alpha') + ':'), walpha, wclear, wquit):
            hbox.addWidget(w, stretch=0)

        hw = QtGui.QWidget()
        hw.setLayout(hbox)
        vbox.addWidget(hw, stretch=0)

        mode = self.canvas.get_draw_mode()
        hbox = QtGui.QHBoxLayout()
        hbox.setContentsMargins(QtCore.QMargins(4, 2, 4, 2))

        btn1 = QtGui.QRadioButton(_("Draw"))
        btn1.setChecked(mode == 'draw')
        btn1.toggled.connect(lambda val: self.set_mode_cb('draw', val))
        btn1.setToolTip(_("Choose this to draw on the canvas"))
        hbox.addWidget(btn1)

        btn2 = QtGui.QRadioButton(_("Edit"))
        btn2.setChecked(mode == 'edit')
        btn2.toggled.connect(lambda val: self.set_mode_cb('edit', val))
        btn2.setToolTip(_("Choose this to edit things on the canvas"))
        hbox.addWidget(btn2)

        btn3 = QtGui.QRadioButton(_("Pick"))
        btn3.setChecked(mode == 'pick')
        btn3.toggled.connect(lambda val: self.set_mode_cb('pick', val))
        btn3.setToolTip(_("Choose this to pick things on the canvas"))
        hbox.addWidget(btn3)

        hbox.addWidget(QtGui.QLabel(''), stretch=1)
        hw = QtGui.QWidget()
        hw.setLayout(hbox)
        vbox.addWidget(hw, stretch=0)

        vw = QtGui.QWidget()
        self.setCentralWidget(vw)
        vw.setLayout(vbox)

    def set_drawparams(self, kind):
        index = self.wdrawtype.currentIndex()
        kind = self.drawtypes[index]
        index = self.wdrawcolor.currentIndex()
        fill = (self.wfill.checkState() != 0)
        alpha = self.walpha.value()

        params = {'color': self.drawcolors[index],
                  'alpha': alpha,
                  }
        if kind in ('circle', 'rectangle', 'polygon', 'triangle',
                    'righttriangle', 'ellipse', 'square', 'box',
                    'squarebox'):
            params['fill'] = fill
            params['fillalpha'] = alpha

        self.canvas.set_drawtype(kind, **params)

    def clear_canvas(self):
        self.canvas.delete_all_objects()

    def load_file(self, filepath):
        image = load_data(filepath, logger=self.logger)

        self.fitsimage.set_image(image)
        self.setWindowTitle(filepath)

    def open_file(self):
        res = QtGui.QFileDialog.getOpenFileName(self, _("Open FITS file"),
                                                ".", "FITS files (*.fits)")
        if isinstance(res, tuple):
            fileName = res[0]
        else:
            fileName = str(res)
        if len(fileName) != 0:
            self.load_file(fileName)

    def drop_file_cb(self, fitsimage, paths):
        fileName = paths[0]
        self.load_file(fileName)

    def cursor_cb(self, viewer, button, data_x, data_y):
        """This gets called when the data position relative to the cursor
        changes.
        """
        # Get the value under the data coordinates
        try:
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = viewer.get_data(int(data_x + viewer.data_off),
                                    int(data_y + viewer.data_off))

        except Exception:
            value = None

        fits_x, fits_y = data_x + 1, data_y + 1

        # Calculate WCS RA
        try:
            # NOTE: image function operates on DATA space coords
            image = viewer.get_image()
            if image is None:
                # No image loaded
                return
            ra_txt, dec_txt = image.pixtoradec(fits_x, fits_y,
                                               format='str', coords='fits')
        except Exception as e:
            self.logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = _('BAD WCS')
            dec_txt = _('BAD WCS')

        text = "RA: %s  DEC: %s  X: %.2f  Y: %.2f  Value: %s" % (
            ra_txt, dec_txt, fits_x, fits_y, value)
        self.readout.setText(text)

    def set_mode_cb(self, mode, tf):
        self.logger.info("canvas mode changed (%s) %s" % (mode, tf))
        if not (tf is False):
            self.canvas.set_draw_mode(mode)
        return True

    def draw_cb(self, canvas, tag):
        obj = canvas.get_object_by_tag(tag)
        obj.add_callback('pick-down', self.pick_cb, 'down')
        obj.add_callback('pick-up', self.pick_cb, 'up')
        obj.add_callback('pick-move', self.pick_cb, 'move')
        obj.add_callback('pick-hover', self.pick_cb, 'hover')
        obj.add_callback('pick-enter', self.pick_cb, 'enter')
        obj.add_callback('pick-leave', self.pick_cb, 'leave')
        obj.add_callback('pick-key', self.pick_cb, 'key')
        obj.pickable = True
        obj.add_callback('edited', self.edit_cb)

    def pick_cb(self, obj, canvas, event, pt, ptype):
        self.logger.info("pick event '%s' with obj %s at (%.2f, %.2f)" % (
            ptype, obj.kind, pt[0], pt[1]))
        return True

    def edit_cb(self, obj):
        self.logger.info("object %s has been edited" % (obj.kind))
        return True

    def quit(self, *args):
        self.logger.info("Attempting to shut down the application...")
        self.deleteLater()


def main(options, args):

    if options.render == 'opengl':
        set_default_opengl_context()

    #QtGui.QApplication.setGraphicsSystem('raster')
    app = QtGui.QApplication(args)

    logger = log.get_logger("example2", options=options)

    w = FitsViewer(logger, render=options.render)
    w.resize(524, 540)
    w.show()
    app.setActiveWindow(w)
    w.raise_()
    w.activateWindow()

    if len(args) > 0:
        w.load_file(args[0])

    app.exec_()


if __name__ == "__main__":

    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser()

    argprs.add_argument("--debug", dest="debug", default=False,
                        action="store_true",
                        help="Enter the pdb debugger on main()")
    argprs.add_argument("-r", "--render", dest="render", default='widget',
                        help="Set render type {widget|scene|opengl}")
    argprs.add_argument("--profile", dest="profile", action="store_true",
                        default=False,
                        help="Run the profiler on main()")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    # Are we debugging this?
    if options.debug:
        import pdb

        pdb.run('main(options, args)')

    # Are we profiling this?
    elif options.profile:
        import profile

        print(("%s profile:" % sys.argv[0]))
        profile.run('main(options, args)')

    else:
        main(options, args)

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
``Toolbar`` provides a set of convenience UI controls for common operations
on viewers.

**Plugin Type: Global**

``Toolbar`` is a global plugin.  Only one instance can be opened.

**Usage**

Hovering over an icon on the toolbar should provide you with usage tool tip.

"""
import os.path

from ginga.gw import Widgets
from ginga.toolkit import family
from ginga.misc import Bunch
from ginga.events import KeyEvent
from ginga import GingaPlugin
from ginga.util import viewer as gviewer

__all__ = ['Toolbar', 'Toolbar_Common', 'Toolbar_Ginga_Image',
           'Toolbar_Ginga_Plot', 'Toolbar_Ginga_Table']


class Toolbar(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super().__init__(fv)

        # active view
        self.channel = None
        self.opname_prefix = 'Toolbar_'

        # get local plugin preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Toolbar')
        self.settings.add_defaults(close_unfocused_toolbars=True)
        self.settings.load(onError='silent')

        fv.add_callback('add-channel', self.add_channel_cb)
        fv.add_callback('delete-channel', self.delete_channel_cb)
        fv.add_callback('channel-change', self.focus_cb)
        fv.add_callback('viewer-select', self.viewer_select_cb)

    def build_gui(self, container):
        self.w.nb = Widgets.StackWidget()

        container.add_widget(self.w.nb, stretch=1)

    # CALLBACKS

    def add_channel_cb(self, fv, channel):
        if self.channel is None:
            self.focus_cb(fv, channel)

    def delete_channel_cb(self, fv, channel):
        self.logger.debug("delete channel %s" % (channel.name))
        if channel is self.channel:
            self.channel = None

    def focus_cb(self, fv, channel):
        self.logger.debug("{} focused".format(channel.name))
        if channel is not self.channel:
            old_channel, self.channel = self.channel, channel

            if self.channel is not None:
                opname = self.get_opname(self.channel)
                self.logger.debug(f"starting {opname} in {channel}")
                self.start_local_plugin(self.channel, opname)

            # NOTE: can stop toolbar plugins that aren't focused
            # but it is more efficient to keep them open
            if self.settings.get('close_unfocused_toolbars', False):
                if old_channel is not None:
                    opname = self.get_opname(old_channel)
                    self.logger.debug(f"stopping {opname} in {old_channel}")
                    self.stop_local_plugin(old_channel, opname)

        return True

    def viewer_select_cb(self, fv, channel, old_viewer, new_viewer):
        self.logger.debug("viewer changed: {}".format(new_viewer.name))
        if channel is self.channel and channel is not None:
            #opname = self.get_opname(self.channel)
            opname = self.opname_prefix + new_viewer.vname.replace(' ', '_')
            self.start_local_plugin(self.channel, opname)

            if old_viewer is not None and old_viewer is not new_viewer:
                opname = self.opname_prefix + old_viewer.vname.replace(' ', '_')
                self.stop_local_plugin(self.channel, opname)

    def redo(self, channel, dataobj):
        # NOTE: we need to call redo() specifically for our toolbars
        # because they are not started nor managed like regular local
        # plugins by the core, they are managed by us
        opname = self.get_opname(channel)
        opmon = channel.opmon
        p_obj = opmon.get_plugin(opname)
        try:
            p_obj.redo()
        except Exception as e:
            self.logger.error(f"error updating toolbar {opmon}: {e}")

    # LOGIC

    def get_opname(self, channel):
        opname = self.opname_prefix + channel.viewer.vname.replace(' ', '_')
        return opname

    def start_local_plugin(self, channel, opname, future=None):
        wname = "{}_{}".format(channel.name, opname)
        if wname in self.w and self.w[wname] is not None:
            vbox = self.w[wname]
            idx = self.w.nb.index_of(vbox)
            self.logger.debug(f"raising {wname}")
            self.w.nb.set_index(idx)
            return
        opmon = channel.opmon
        p_obj = opmon.get_plugin(opname)
        vbox = Widgets.VBox()
        p_obj.build_gui(vbox)
        self.w.nb.add_widget(vbox)
        self.w[wname] = vbox
        p_obj.start()

    def stop_local_plugin(self, channel, opname):
        opmon = channel.opmon
        p_obj = opmon.get_plugin(opname)
        try:
            p_obj.stop()
        except Exception as e:
            pass
        wname = "{}_{}".format(channel.name, opname)
        try:
            vbox = self.w[wname]
            self.w.nb.remove(vbox)
            self.w[wname] = None
            vbox.delete()
        except Exception as e:
            self.logger.error(f"error stopping plugin '{wname}'")

    def __str__(self):
        return 'toolbar'


class Toolbar_Common(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super().__init__(fv, fitsimage)

        # holds our gui widgets
        self.w = Bunch.Bunch()
        self.viewers = []
        self.layout_common = [
            # (Name, type, icon, tooltip)
            ("viewer", 'combobox', None, "Select compatible viewer",
             self.viewer_cb),
            ("Up", 'button', 'up', "Go to previous image in channel",
             lambda w: self.fv.prev_img()),
            ("Down", 'button', 'down', "Go to next image in channel",
             lambda w: self.fv.next_img()),
            ("---",)]

        self.gui_up = False

    def close(self):
        self.fv.stop_local_plugin(self.chname, str(self))

    def stop(self):
        self.viewers = []
        self.gui_up = False

    def start_local_plugin(self, name):
        self.fv.start_operation(name)

    def start_global_plugin(self, name):
        self.fv.start_global_plugin(name, raise_tab=True)

    def build_toolbar(self, tb_w, layout):

        for tup in layout:
            name = tup[0]
            if name == '---':
                tb_w.add_separator()
                continue
            if tup[1] not in ['button', 'toggle']:
                btn = Widgets.make_widget(name, tup[1])
                tb_w.add_widget(btn)
            else:
                iconpath = tup[2]
                if not iconpath.startswith(os.path.sep):
                    iconpath = os.path.join(self.fv.iconpath, "%s.svg" % (iconpath))
                btn = tb_w.add_action(None, toggle=(tup[1] == 'toggle'),
                                      iconpath=iconpath, iconsize=(24, 24))
            if tup[3]:
                btn.set_tooltip(tup[3])
            if tup[4]:
                btn.add_callback('activated', tup[4])

            # add to our widget dict
            self.w[Widgets.name_mangle(name, pfx='btn_')] = btn

        return tb_w

    def viewer_cb(self, w, idx):
        vinfo = self.viewers[idx]
        self.logger.debug(f"viewer {vinfo.name} selected")
        dataobj = self.channel.viewer.get_dataobj()
        self.channel.open_with_viewer(vinfo, dataobj)

    def _update_viewer_selection(self):
        if not self.gui_up:
            return

        dataobj = self.channel.get_current_image()

        # find available viewers that can view this kind of object
        viewers = gviewer.get_viewers(dataobj)
        if viewers != self.viewers:
            # repopulate viewer selector
            self.viewers = viewers
            new_names = [viewer.name for viewer in viewers]
            self.w.btn_viewer.clear()
            self.logger.debug("available viewers for {} are {}".format(type(dataobj), new_names))
            for name in new_names:
                self.w.btn_viewer.append_text(name)
            # set the box to the viewer we have selected
            cur_name = self.channel.viewer.vname
            if cur_name in new_names:
                self.w.btn_viewer.set_text(cur_name)

    def __str__(self):
        return 'toolbarbase'


class Toolbar_Ginga_Image(Toolbar_Common):
    """A toolbar for the Ginga Image viewer.
    """

    def __init__(self, fv, chviewer):
        # superclass defines some variables for us, like logger
        super().__init__(fv, chviewer)

        # get local plugin preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('toolbar_Ginga_Image')
        self.settings.load(onError='silent')

        self.modetype = self.settings.get('mode_type', 'oneshot')

        chviewer.add_callback('transform', self.viewer_transform_cb)

        bm = chviewer.get_bindmap()
        bm.add_callback('mode-set', self.mode_set_cb)

        self.layout = [
            # (Name, type, icon, tooltip)
            ("FlipX", 'toggle', 'flip_x', "Flip image in X axis",
             self.flipx_cb),
            ("FlipY", 'toggle', 'flip_y', "Flip image in Y axis",
             self.flipy_cb),
            ("SwapXY", 'toggle', 'swap_xy', "Swap X and Y axes",
             self.swapxy_cb),
            ("---",),
            ("Rot90", 'button', 'rot90ccw', "Rotate image 90 deg",
             self.rot90_cb),
            ("RotN90", 'button', 'rot90cw', "Rotate image -90 deg",
             self.rotn90_cb),
            ("OrientRH", 'button', 'orient_nw', "Orient image N=Up E=Right",
             self.orient_rh_cb),
            ("OrientLH", 'button', 'orient_ne', "Orient image N=Up E=Left",
             self.orient_lh_cb),
            ("---",),
            ("Zoom In", 'button', 'zoom_in', "Zoom in",
             lambda w: self.fv.zoom_in()),
            ("Zoom Out", 'button', 'zoom_out', "Zoom out",
             lambda w: self.fv.zoom_out()),
            ("Zoom Fit", 'button', 'zoom_fit', "Zoom to fit window size",
             lambda w: self.fv.zoom_fit()),
            ("Zoom 1:1", 'button', 'zoom_100', "Zoom to 100% (1:1)",
             lambda w: self.fv.zoom_1_to_1()),
            ("---",),
            ("Pan", 'toggle', 'pan', "Pan with left, zoom with right",
             lambda w, tf: self.mode_cb(tf, 'pan')),
            ("Zoom", 'toggle', 'crosshair',
             "Left/right click zooms in/out;\n"
             "hold middle to pan freely over image",
             lambda w, tf: self.mode_cb(tf, 'zoom')),
            ("Rotate", 'toggle', 'rotate',
             "Drag left to rotate; click right to reset to 0 deg",
             lambda w, tf: self.mode_cb(tf, 'rotate')),
            ("Dist", 'toggle', 'sqrt',
             "Scroll to set color distribution algorithm",
             lambda w, tf: self.mode_cb(tf, 'dist')),
            ("CMap", 'toggle', 'palette',
             "Scroll to set color map",
             lambda w, tf: self.mode_cb(tf, 'cmap')),
            ("Cuts", 'toggle', 'cuts',
             "Left/right sets hi cut, up/down sets lo cut",
             lambda w, tf: self.mode_cb(tf, 'cuts')),
            ("Contrast", 'toggle', 'contrast',
             "Contrast/bias with left/right/up/down",
             lambda w, tf: self.mode_cb(tf, 'contrast')),
            ("ModeLock", 'toggle', 'lock',
             "Modes are oneshot or locked", self.set_locked_cb),
            ("---",),
            ("Center", 'button', 'center_image', "Center image",
             self.center_image_cb),
            ("Restore", 'button', 'reset_rotation',
             "Reset all transformations and rotations",
             self.reset_all_transforms_cb),
            ("AutoLevels", 'button', 'auto_cuts', "Auto cut levels",
             self.auto_levels_cb),
            ("ResetContrast", 'button', 'reset_contrast', "Reset contrast",
             self.reset_contrast_cb),
            ("---",),
            ("Preferences", 'button', 'settings', "Set channel preferences (in focused channel)",
             lambda w: self.start_local_plugin('Preferences')),
            ("FBrowser", 'button', 'folder_open', "Open file (in focused channel)",
             lambda w: self.start_local_plugin('FBrowser')),
            ("MultiDim", 'button', 'layers', "Select HDUs or cube slices (in focused channel)",
             lambda w: self.start_local_plugin('MultiDim')),
            ("Header", 'button', 'tags', "View image metadata (Header plugin)",
             lambda w: self.start_global_plugin('Header')),
            ("ZoomPlugin", 'button', 'microscope', "Magnify detail (Zoom plugin)",
             lambda w: self.start_global_plugin('Zoom'))]

    def build_gui(self, container):
        # TODO: fix for GTK
        if family.startswith('gtk'):
            self.orientation = 'horizontal'
        else:
            self.orientation = Widgets.get_orientation(container)
        tb_w = Widgets.Toolbar(orientation=self.orientation)

        self.build_toolbar(tb_w, self.layout_common)
        self.build_toolbar(tb_w, self.layout)
        self.w.toolbar = tb_w

        container.add_widget(tb_w, stretch=1)
        self.gui_up = True

    def start(self):
        self._update_toolbar_state()

    # CALLBACKS

    def center_image_cb(self, w):
        self.fitsimage.center_image()
        return True

    def reset_contrast_cb(self, w):
        view, mode = self._get_mode('contrast')
        event = KeyEvent()
        mode.kp_contrast_restore(view, event, 'x', 0.0, 0.0)
        return True

    def auto_levels_cb(self, w):
        self.fitsimage.auto_levels()
        return True

    def rot90_cb(self, w):
        self.fitsimage.rotate_delta(90.0)
        return True

    def rotn90_cb(self, w):
        self.fitsimage.rotate_delta(-90.0)
        return True

    def orient_lh_cb(self, w):
        view, mode = self._get_mode('rotate')
        event = KeyEvent()
        mode.kp_orient_lh(view, event, 'x', 0.0, 0.0)
        return True

    def orient_rh_cb(self, w):
        view, mode = self._get_mode('rotate')
        event = KeyEvent()
        mode.kp_orient_rh(view, event, 'x', 0.0, 0.0)
        return True

    def reset_all_transforms_cb(self, w):
        view = self.fitsimage
        with view.suppress_redraw:
            view.rotate(0.0)
            view.transform(False, False, False)
        return True

    def flipx_cb(self, w, tf):
        view = self.fitsimage
        flip_x, flip_y, swap_xy = view.get_transforms()
        flip_x = tf
        view.transform(flip_x, flip_y, swap_xy)
        return True

    def flipy_cb(self, w, tf):
        view = self.fitsimage
        flip_x, flip_y, swap_xy = view.get_transforms()
        flip_y = tf
        view.transform(flip_x, flip_y, swap_xy)
        return True

    def swapxy_cb(self, w, tf):
        view = self.fitsimage
        flip_x, flip_y, swap_xy = view.get_transforms()
        swap_xy = tf
        view.transform(flip_x, flip_y, swap_xy)
        return True

    def mode_cb(self, tf, modename):
        bm = self.fitsimage.get_bindmap()
        if not tf:
            bm.reset_mode(self.fitsimage)
            self.fv.show_status("")
            return True

        bm.set_mode(modename)
        # just in case mode change failed
        self._update_toolbar_state()
        self.fv.show_status(f"Type 'h' in the viewer to show help for mode {modename}")
        return True

    def mode_set_cb(self, bm, mode, mtype):
        # called whenever the user interaction mode is changed
        # in the viewer
        self._update_toolbar_state()
        return True

    def viewer_transform_cb(self, chviewer):
        # called whenever the transform (flip x/y, swap axes) is done
        # in the viewer
        self._update_toolbar_state()
        return True

    def set_locked_cb(self, w, tf):
        if tf:
            modetype = 'locked'
        else:
            modetype = 'oneshot'
        chviewer = self.fitsimage

        # get current bindmap, make sure that the mode is consistent
        # with current lock button
        bm = chviewer.get_bindmap()
        modename, cur_modetype = bm.current_mode()
        bm.set_default_mode_type(modetype)

        bm.set_mode(modename, mode_type=modetype)
        if not tf:
            # turning off lock also resets the mode
            bm.reset_mode(chviewer)

        self._update_toolbar_state()
        return True

    # LOGIC

    def _get_mode(self, mode_name):
        view = self.fitsimage
        bd = view.get_bindings()
        mode = bd.get_mode_obj(mode_name)
        return view, mode

    def redo(self):
        self._update_toolbar_state()

    def clear(self):
        self._update_toolbar_state()

    def _update_toolbar_state(self):
        if not self.gui_up:
            return
        self.logger.debug("updating toolbar state")
        chviewer = self.fitsimage
        try:
            self._update_viewer_selection()

            # Update toolbar channel buttons
            enabled = len(self.channel) > 1
            self.w.btn_up.set_enabled(enabled)
            self.w.btn_down.set_enabled(enabled)

            bd = chviewer.get_bindings()
            can_flip = bd.get_feature_allow('flip')
            self.w.btn_flipx.set_enabled(can_flip)
            self.w.btn_flipy.set_enabled(can_flip)
            self.w.btn_swapxy.set_enabled(can_flip)

            # update transform toggles
            flipx, flipy, swapxy = chviewer.get_transforms()
            # toolbar follows view
            self.w.btn_flipx.set_state(flipx)
            self.w.btn_flipy.set_state(flipy)
            self.w.btn_swapxy.set_state(swapxy)

            can_rotate = bd.get_feature_allow('rotate')
            self.w.btn_rotate.set_enabled(can_rotate)
            self.w.btn_restore.set_enabled(can_rotate & can_flip)
            self.w.btn_rot90.set_enabled(can_rotate)
            self.w.btn_rotn90.set_enabled(can_rotate)
            self.w.btn_orientrh.set_enabled(can_rotate)
            self.w.btn_orientlh.set_enabled(can_rotate)

            can_pan = bd.get_feature_allow('pan')
            can_zoom = bd.get_feature_allow('zoom')
            self.w.btn_pan.set_enabled(can_pan & can_zoom)
            self.w.btn_center.set_enabled(can_pan)

            self.w.btn_zoom_in.set_enabled(can_zoom)
            self.w.btn_zoom_out.set_enabled(can_zoom)
            self.w.btn_zoom_fit.set_enabled(can_zoom)
            self.w.btn_zoom_1_1.set_enabled(can_zoom)
            self.w.btn_zoom.set_enabled(can_pan & can_zoom)

            can_cut = bd.get_feature_allow('cut')
            self.w.btn_cuts.set_enabled(can_cut)
            self.w.btn_autolevels.set_enabled(can_cut)

            can_cmap = bd.get_feature_allow('cmap')
            self.w.btn_contrast.set_enabled(can_cmap)
            self.w.btn_resetcontrast.set_enabled(can_cmap)
            self.w.btn_dist.set_enabled(can_cmap)
            self.w.btn_cmap.set_enabled(can_cmap)

            # update mode toggles
            bm = chviewer.get_bindmap()
            modename, mode_type = bm.current_mode()
            self.logger.debug("modename=%s" % (modename))
            # toolbar follows view
            self.w.btn_pan.set_state(modename == 'pan')
            self.w.btn_zoom.set_state(modename == 'zoom')
            self.w.btn_rotate.set_state(modename == 'rotate')
            self.w.btn_dist.set_state(modename == 'dist')
            self.w.btn_cuts.set_state(modename == 'cuts')
            self.w.btn_contrast.set_state(modename == 'contrast')
            self.w.btn_cmap.set_state(modename == 'cmap')

            default_mode_type = bm.get_default_mode_type()
            if 'btn_modelock' in self.w:
                is_locked = (default_mode_type in ('locked', 'softlock'))
                self.w.btn_modelock.set_state(is_locked)

        except Exception as e:
            self.logger.error("error updating toolbar: %s" % str(e))
            raise e

    def __str__(self):
        return 'toolbar_ginga_image'


class Toolbar_Ginga_Plot(Toolbar_Common):
    """A toolbar for the Ginga Plot viewer.
    """

    def __init__(self, fv, chviewer):
        # superclass defines some variables for us, like logger
        super().__init__(fv, chviewer)

        # get local plugin preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('toolbar_Ginga_Plot')
        self.settings.add_defaults(zoom_bump_pct=1.1)
        self.settings.load(onError='silent')

        self.layout = [
            # (Name, type, icon, tooltip)
            ("Zoom In", 'button', 'zoom_in', "Zoom in",
             lambda w: self.zoom_in()),
            ("Zoom Out", 'button', 'zoom_out', "Zoom out",
             lambda w: self.zoom_out()),
            ("Zoom Fit", 'button', 'zoom_fit', "Fit plot to window size",
             lambda w: self.zoom_fit()),
            ("Zoom Fit X", 'button', 'zoom_fit_x', "Fit X axis to window size",
             lambda w: self.zoom_fit_x()),
            ("Zoom Fit Y", 'button', 'zoom_fit_y', "Fit Y axis to window size",
             lambda w: self.zoom_fit_y()),
            ("---",),
            ("Preferences", 'button', 'settings', "Set channel preferences (in focused channel)",
             lambda w: self.start_local_plugin('Preferences')),
            ("FBrowser", 'button', 'folder_open', "Open file (in focused channel)",
             lambda w: self.start_local_plugin('FBrowser')),
            ("MultiDim", 'button', 'layers', "Select HDUs or cube slices (in focused channel)",
             lambda w: self.start_local_plugin('MultiDim')),
            ("Header", 'button', 'tags', "View image metadata (Header plugin)",
             lambda w: self.start_global_plugin('Header'))]

    def build_gui(self, container):
        # TODO: fix for GTK
        if family.startswith('gtk'):
            self.orientation = 'horizontal'
        else:
            self.orientation = Widgets.get_orientation(container)
        tb_w = Widgets.Toolbar(orientation=self.orientation)

        self.build_toolbar(tb_w, self.layout_common)
        self.build_toolbar(tb_w, self.layout)
        self.w.toolbar = tb_w

        container.add_widget(tb_w, stretch=1)
        self.gui_up = True

    def start(self):
        self._update_toolbar_state()

    # CALLBACKS

    def zoom_in(self):
        viewer = self.channel.get_viewer('Ginga Plot')
        pct = self.settings['zoom_bump_pct']
        viewer.zoom_plot(1 / pct, 1 / pct)

    def zoom_out(self):
        viewer = self.channel.get_viewer('Ginga Plot')
        pct = self.settings['zoom_bump_pct']
        viewer.zoom_plot(pct, pct)

    def zoom_fit_x(self):
        viewer = self.channel.get_viewer('Ginga Plot')
        viewer.zoom_fit(axis='x')

    def zoom_fit_y(self):
        viewer = self.channel.get_viewer('Ginga Plot')
        viewer.zoom_fit(axis='y')

    def zoom_fit(self):
        viewer = self.channel.get_viewer('Ginga Plot')
        viewer.zoom_fit(axis='xy')

    # LOGIC

    def redo(self):
        self._update_toolbar_state()

    def clear(self):
        self._update_toolbar_state()

    def _update_toolbar_state(self):
        if not self.gui_up:
            return
        self.logger.debug("updating toolbar state")
        try:
            self._update_viewer_selection()

            # Update toolbar channel buttons
            enabled = len(self.channel) > 1
            self.w.btn_up.set_enabled(enabled)
            self.w.btn_down.set_enabled(enabled)

        except Exception as e:
            self.logger.error("error updating toolbar: %s" % str(e))
            raise e

    def __str__(self):
        return 'toolbar_ginga_plot'


class Toolbar_Ginga_Table(Toolbar_Common):
    """A toolbar for the Ginga Table viewer.
    """

    def __init__(self, fv, chviewer):
        # superclass defines some variables for us, like logger
        super().__init__(fv, chviewer)

        # get local plugin preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('toolbar_Ginga_Table')
        self.settings.load(onError='silent')

        self.layout = [
            # (Name, type, icon, tooltip)
            ("Preferences", 'button', 'settings', "Set channel preferences (in focused channel)",
             lambda w: self.start_local_plugin('Preferences')),
            ("FBrowser", 'button', 'folder_open', "Open file (in focused channel)",
             lambda w: self.start_local_plugin('FBrowser')),
            ("MultiDim", 'button', 'layers', "Select HDUs or cube slices (in focused channel)",
             lambda w: self.start_local_plugin('MultiDim')),
            ("Header", 'button', 'tags', "View image metadata (Header plugin)",
             lambda w: self.start_global_plugin('Header'))]

    def build_gui(self, container):
        # TODO: fix for GTK
        if family.startswith('gtk'):
            self.orientation = 'horizontal'
        else:
            self.orientation = Widgets.get_orientation(container)
        tb_w = Widgets.Toolbar(orientation=self.orientation)

        self.build_toolbar(tb_w, self.layout_common)
        self.build_toolbar(tb_w, self.layout)
        self.w.toolbar = tb_w

        container.add_widget(tb_w, stretch=1)
        self.gui_up = True

    def start(self):
        self._update_toolbar_state()

    # LOGIC

    def redo(self):
        self._update_toolbar_state()

    def clear(self):
        self._update_toolbar_state()

    def _update_toolbar_state(self):
        if not self.gui_up:
            return
        self.logger.debug("updating toolbar state")
        try:
            self._update_viewer_selection()

            # Update toolbar channel buttons
            enabled = len(self.channel) > 1
            self.w.btn_up.set_enabled(enabled)
            self.w.btn_down.set_enabled(enabled)

        except Exception as e:
            self.logger.error("error updating toolbar: %s" % str(e))
            raise e

    def __str__(self):
        return 'toolbar_ginga_table'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Toolbar', package='ginga')

# END



#
# Plotable.py -- Abstraction of generic plot data.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.BaseImage import ViewerObjectBase, Header
from ginga.canvas.CanvasObject import get_canvas_types


class Plotable(ViewerObjectBase):
    """Abstraction of a plot data.

    .. note:: This module is NOT thread-safe!

    """
    # class variables for WCS can be set
    wcsClass = None
    ioClass = None

    @classmethod
    def set_wcsClass(cls, klass):
        cls.wcsClass = klass

    @classmethod
    def set_ioClass(cls, klass):
        cls.ioClass = klass

    def __init__(self, data_np=None, metadata=None, logger=None, name=None,
                 wcsclass=wcsClass, ioclass=ioClass):

        ViewerObjectBase.__init__(self, logger=logger, metadata=metadata,
                                  name=name)

        self.wcs = None
        self.io = None

        self.dc = get_canvas_types()
        self.canvas = self.dc.Canvas()
        self.set_defaults(title=None, grid=False, legend=False, clr_bg='white',
                          x_axis_label=None, y_axis_label=None,
                          x_axis_scale='linear', y_axis_scale='linear')
        self.rgb_order = 'RGBA'

        if data_np is not None:
            self.plot_line(data_np)

    def get_size(self):
        return (self.columns, self.rows)

    def get_canvas(self):
        return self.canvas

    def get_header(self, create=True, include_primary_header=None):
        # By convention, the header is stored in a dictionary
        # under the metadata keyword 'header'
        if 'header' not in self:
            if not create:
                # TODO: change to ValueError("No header found")
                raise KeyError('header')

            hdr = Header()
            self.set(header=hdr)
        else:
            hdr = self['header']

        return hdr

    def has_primary_header(self):
        return False

    def set_titles(self, title=None, x_axis=None, y_axis=None):
        if x_axis is not None:
            self.set(x_axis_label=x_axis)
        if y_axis is not None:
            self.set(y_axis_label=y_axis)
        if title is not None:
            self.set(title=title)

    def set_grid(self, tf):
        self.set(grid=tf)

    def set_legend(self, tf):
        self.set(legend=tf)

    def set_bg(self, color):
        self.set(clr_bg=color)

    def set_axes_scale(self, x_axis=None, y_axis=None):
        kwargs = dict()
        if x_axis is not None:
            kwargs['x_axis_scale'] = x_axis
        if y_axis is not None:
            kwargs['y_axis_scale'] = y_axis
        if len(kwargs) > 0:
            self.set(**kwargs)

    def clear(self):
        self.canvas.delete_all_objects()
        self.set(x_axis_label=None, y_axis_label=None, title=None,
                 grid=False, legend=False)

        self.make_callback('modified')

    def get_minmax(self, noinf=False):
        # TODO: what should this mean for a plot?
        return (0, 0)

    def load_file(self, filespec):
        raise NotImplementedError("This method is not yet implemented")

    def get_thumbnail(self, length):
        thumb_np = np.eye(length)
        return thumb_np

    def plot_line(self, points, color='black', linewidth=1,
                  alpha=1.0, name=None, tag=None):
        """Simple method to plot a line."""
        path = self.dc.Path(points, color=color, linewidth=linewidth,
                            name=name)
        self.canvas.add(path, tag=tag)

        self.make_callback('modified')

    def info_xy(self, data_x, data_y, settings):
        info = super().info_xy(data_x, data_y, settings)

        ra_txt = "%+.3f" % (data_x)
        dec_txt = "%+.3f" % (data_y)
        ra_lbl, dec_lbl = "X", "Y"

        info.update(dict(itype='plotxy', ra_txt=ra_txt, dec_txt=dec_txt,
                         ra_lbl=ra_lbl, dec_lbl=dec_lbl))
        return info



#
# CanvasRenderCv.py -- for rendering into a Ginga widget with OpenCv
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import numpy as np

import cv2  # noqa
from . import CvHelp

from ginga.canvas import render
# force registration of all canvas types
import ginga.canvas.types.all  # noqa
from ginga import trcalc


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)

        # special scaling for Cv text drawing to normalize it relative
        # to other backends
        self._font_scale_factor = 1.33

        # TODO: encapsulate this drawable
        self.ctx = CvHelp.CvContext(surface)

    def get_line(self, color, alpha=1.0, linewidth=1, linestyle='solid'):
        line = super().get_line(color, alpha=alpha, linewidth=linewidth,
                                linestyle=linestyle)

        line.render.color = line.get_bpp_color(8)
        return line

    def get_fill(self, color, alpha=1.0):
        fill = super().get_fill(color, alpha=alpha)

        fill.render.color = fill.get_bpp_color(8)
        return fill

    def get_font(self, fontname, **kwargs):
        font = super().get_font(fontname, **kwargs)

        font.render.scale = int(round(font.fontsize * self._font_scale_factor))
        font.render.font = CvHelp.get_cached_font(font.fontname, font.fontsize)
        return font

    def text_extents(self, text, font=None):
        if font is None:
            font = self.font
        _font, _scale = font.render.font, font.render.scale
        retval, baseline = _font.getTextSize(text, _scale, -1)
        wd_px, ht_px = retval
        return wd_px, ht_px

    ##### DRAWING OPERATIONS #####

    def draw_image(self, cvs_img, cpoints, rgb_arr, whence, order='RGBA'):
        # no-op for this renderer
        pass

    def draw_text(self, cx, cy, text, rot_deg=0.0, font=None, fill=None,
                  line=None):
        self.ctx.text((cx, cy), text, font, line, fill)

    def draw_polygon(self, cpoints, line=None, fill=None):
        cpoints = trcalc.strip_z(cpoints)
        self.ctx.polygon(cpoints, line, fill)

    def draw_circle(self, cx, cy, cradius, line=None, fill=None):
        self.ctx.circle((cx, cy), cradius, line, fill)

    def draw_line(self, cx1, cy1, cx2, cy2, line=None):
        self.ctx.line((cx1, cy1), (cx2, cy2), line)

    def draw_path(self, cpoints, line=None):
        cpoints = trcalc.strip_z(cpoints)
        self.ctx.path(cpoints, line)


class CanvasRenderer(render.StandardPipelineRenderer):

    def __init__(self, viewer):
        render.StandardPipelineRenderer.__init__(self, viewer)

        self.kind = 'opencv'
        # According to OpenCV documentation:
        # "If you are using your own image rendering and I/O functions,
        # you can use any channel ordering. The drawing functions process
        # each channel independently and do not depend on the channel
        # order or even on the used color space."
        # NOTE: OpenCv does not seem to be happy using anti-aliasing on
        # transparent arrays
        self.rgb_order = 'RGB'
        self.surface = None
        self.dims = ()

    def resize(self, dims):
        """Resize our drawing area to encompass a space defined by the
        given dimensions.
        """
        width, height = dims[:2]
        self.logger.debug("renderer reconfigured to %dx%d" % (
            width, height))

        # create cv surface the size of the window
        # (cv just uses numpy arrays!)
        depth = len(self.rgb_order)
        self.surface = np.zeros((height, width, depth), dtype=np.uint8)

        super(CanvasRenderer, self).resize(dims)

    def render_image(self, rgb_arr, order, win_coord):
        """Render the image represented by (rgb_arr) at (win_coord)
        in the pixel space.
        *** internal method-- do not use ***
        """
        if self.surface is None:
            return
        self.logger.debug("redraw surface")

        # TODO: is there a faster way to copy this array in?
        self.surface[:, :, :] = rgb_arr[:, :, 0:3]

    def get_surface_as_array(self, order=None):
        if self.surface is None:
            raise render.RenderError("No OpenCv surface defined")

        # adjust according to viewer's needed order
        return self.reorder(order, self.surface)

    def setup_cr(self, shape):
        cr = RenderContext(self, self.viewer, self.surface)
        return cr

    def get_dimensions(self, shape):
        cr = self.setup_cr(shape)
        cr.set_font_from_shape(shape)
        return cr.text_extents(shape.text)

    def text_extents(self, text, font):
        cr = RenderContext(self, self.viewer, self.surface)
        cr.set_font(font.fontname, font.fontsize)
        return cr.text_extents(text)

#END
